import os
import copy
from .storage import PatchVersion
from .binfile import BinFile, BinEmbedded
from .rstfile import RstFile
from .tools import json_dump, stringtable_paths


def load_translations(path, rsthash_version=1415):
    with open(path, "rb") as f:
        if f.read(3) == b"RST":
            f.seek(0)
            return RstFile(f, rsthash_version)
        else:
            translations = {}
            for line in f:
                if line.startswith(b'tr "'):
                    line = line.decode()
                    key, val = line.split("=", 1)
                    key = key[4:-2]
                    val = val[2:-2]
                    translations[key] = val
            return translations


class TftTransformer:
    def __init__(self, input_dir, game_version=1415):
        self.input_dir = input_dir
        self.rsthash_version = game_version

    def build_template(self):
        """Parse bin data into template data"""
        map22_file = os.path.join(self.input_dir, "data", "maps", "shipping", "map22", "map22.bin")

        map22 = BinFile(map22_file)

        character_names = self.parse_character_names(map22)
        traits = self.parse_traits(map22)
        sets = self.parse_sets(map22, character_names, traits)
        champs = self.parse_champs(map22, traits)
        output_sets, output_set_data = self.build_output_sets(sets, champs)
        items = self.parse_items(map22)

        return {
            "sets": output_sets,
            "setData": output_set_data,
            "items": items,
        }

    def export(self, output, langs=None):
        """Export TFT data for given languages

        By default (`langs` is `None`), export all available languages.
        Otherwise, export for given `xx_yy` language codes.
        """

        stringtables = stringtable_paths(self.input_dir, "tft")

        if langs is None:
            langs = list(stringtables)

        os.makedirs(output, exist_ok=True)

        template = self.build_template()
        for lang in langs:
            instance = copy.deepcopy(template)
            replacements = load_translations(stringtables[lang], self.rsthash_version)

            def replace_in_data(entry):
                for key in ("name", "desc"):
                    if key in entry and entry[key] in replacements:
                        entry[key] = replacements[entry[key]]
            def replace_list(l):
                l[:] = [replacements.get(v, v) for v in l]

            def replace_set_data(entry):
                for trait_data in entry["traits"]:
                    replace_in_data(trait_data)
                for champ_data in entry["champions"]:
                    replace_in_data(champ_data)
                    replace_list(champ_data["traits"])
                    if "ability" in champ_data:
                        replace_in_data(champ_data["ability"])

            for set_data in instance["sets"].values():
                replace_set_data(set_data)
            for set_data in instance["setData"]:
                replace_set_data(set_data)
            for data in instance["items"]:
                replace_in_data(data)

            with open(os.path.join(output, f"{lang}.json"), "w", encoding="utf-8") as f:
                json_dump(instance, f, indent=4, sort_keys=True, ensure_ascii=False)

    def build_output_sets(self, sets, champs):
        """Build sets as output in the final JSON file"""

        output_sets = {}
        output_set_data = []
        for set_number, set_mutator, set_name, set_chars, set_traits, set_augments, set_items in sets:
            set_champs_pairs = [champs[name] for name in set_chars if name in champs]
            set_champions = [p[0] for p in set_champs_pairs]
            output_set_data.append({
                "number": set_number,
                "mutator": set_mutator,
                "name": set_name,
                "traits": set_traits,
                "champions": set_champions,
                "augments": set_augments,
                "items": set_items,
            })
            output_sets[set_number] = {
                "name": set_name,
                "traits": set_traits,
                "champions": set_champions,
            }
        return output_sets, output_set_data

    def parse_character_names(self, map22):
        """Parse character names, indexed by entry path"""
        # always use lowercased name: required for files, and bin data is inconsistent
        return {x.path: x.getv("name").lower() for x in map22.entries if x.type == "Character" or x.type == "TftCharacter"}

    def parse_sets(self, map22, character_names, traits):
        """Parse character sets to a list of `(name, number, characters, traits, augments, items)`"""
        character_lists = {x.path: x for x in map22.entries if x.type == "MapCharacterList" or x.type == "TftCharacterList"}
        trait_lists = {x.path: x for x in map22.entries if x.type == "TftTraitList"}
        set_collection = [x for x in map22.entries if x.type == 0x438850FF]
        item_lists = {x.path: x for x in map22.entries if x.type == "TFTItemList"}
        item_entries = {x.path: x for x in map22.entries if x.type == "TftItemData"}

        if not set_collection:
            # backward compatibility
            longest_character_list = max(character_lists.values(), key=lambda v: len(v.fields[0].value))
            longest_trait_list = max(trait_lists.values(), key=lambda v: len(v.fields[0].value))
            champion_list = longest_character_list.fields[0].value
            trait_list = longest_trait_list.fields[0].value
            set_characters = [character_names.get(char) for char in champion_list]
            set_traits = [traits.get(trait) for trait in trait_list]
            return [(1, "Base", set_characters, set_traits)]

        sets = []
        for item in set_collection:
            set_number = item.getv("number")
            set_mutator = item.getv("Mutator")
            if set_mutator is None:
                set_mutator = item.getv("name")
            char_lists = item.getv("characterLists", []) + item.getv("tftCharacterLists", [])
            if not char_lists:
                continue
            set_trait_lists = item.getv("TraitLists")
            if set_trait_lists is None:
                continue
            set_info = item[0xD2538E5A].value
            set_name = set_info["SetName"].getv("mValue")

            if set_number is None or set_name is None:
                continue
            set_characters = []
            for char_list in char_lists:
                if char_list not in character_lists:
                    continue
                set_characters += [character_names[char] for char in character_lists[char_list].getv("Characters") if char in character_names]

            set_trait_paths = []
            for trait_list in set_trait_lists:
                if trait_list not in trait_lists:
                    continue
                set_trait_paths += trait_lists[trait_list].getv("mTraits")
            set_traits = [traits[path] for path in set(set_trait_paths) if path in traits]

            set_augment_ids = set()
            set_items_ids = set()
            for item_list in item.getv("ItemLists", []):
                if item_list not in item_lists:
                    continue
                item_list_entry = item_lists[item_list]
                for item_entry in item_list_entry.getv("mItems", []):
                    if item_entry not in item_entries:
                        continue
                    if item_entries[item_entry].getv("IsAugment"):
                        set_augment_ids.add(item_entries[item_entry].getv("mName"))
                    else:
                        set_items_ids.add(item_entries[item_entry].getv("mName"))
            set_augments = sorted(set_augment_ids)
            set_items = sorted(set_items_ids)

            sets.append((set_number, set_mutator, set_name, set_characters, set_traits, set_augments, set_items))
        return sets

    def parse_items(self, map22):
        item_entries = [x for x in map22.entries if x.type == "TftItemData"]
        trait_entries = [x for x in map22.entries if x.type == "TftTraitData"]

        traits_by_hash = {trait.path.h: trait.getv("mName") for trait in trait_entries}

        items = []
        items_by_hash = {}  # {item_hash: item}
        for item in item_entries:
            name = item.getv("mName")
            if "Template" in name or name == "TFT_Item_Null":
                continue

            effects = {}
            for effect in item.getv("effectAmounts", []):
                name = str(effect.getv("name")) if "name" in effect else effect.path.hex()
                effects[name] = effect.getv("value", "null")

            item_data = {
                "id": item.getv("mId"),
                "name": item.getv(0xC3143D66),
                "apiName": item.getv("mName"),
                "desc": item.getv(0x765F18DA),
                "icon": item.getv("mIconPath"),
                "unique": item.getv(0x9596A387, False),
                "composition": [x.h for x in item.getv(0x8B83BA8A, [])],  # updated below
                "associatedTraits": [x.h for x in item.getv("AssociatedTraits", [])], # updated below
                "incompatibleTraits": [x.h for x in item.getv("IncompatibleTraits", [])], # updated below
                "effects": effects,
            }
            items.append(item_data)
            items_by_hash[item.path.h] = item_data


        for item in items:
            if item["id"] is not None:
                # patchs < 13.5: mId exist and "from" is a list of those IDs
                item["from"] = [items_by_hash[h]["id"] for h in item["composition"]]
            else:
                item["from"] = None
            item["composition"] = [items_by_hash[h]["apiName"] for h in item["composition"]]
            item["incompatibleTraits"] = [traits_by_hash[h] for h in item["incompatibleTraits"]]
            item["associatedTraits"] = [traits_by_hash[h] for h in item["associatedTraits"]]
        return items

    def parse_champs(self, map22, traits):
        """Parse champion information

        Return a map of `(data, traits)`, indexed by champion internal names.
        """
        champ_entries = [x for x in map22.entries if x.type == "TftShopData"]
        champs = {}
        data_characters_dir = os.path.join(self.input_dir, "data", "characters")
        characters_dir = os.path.join(self.input_dir, "characters")

        for champ in champ_entries:
            # always use lowercased name: required for files, and bin data is inconsistent
            name = champ.getv("mName").lower()
            if name == "tft_template":
                continue

            self_path = os.path.join(characters_dir, name + ".cdtb.bin")
            if not os.path.exists(self_path):
                self_path = os.path.join(data_characters_dir, name, name + ".bin")
            if not os.path.exists(self_path):
                continue

            tft_bin = BinFile(self_path)
            record = next((x for x in tft_bin.entries if x.type == "TFTCharacterRecord"), {})
            if "spellNames" not in record:
                continue

            champ_traits = []  # trait paths, as hashes
            for trait in record.getv("mLinkedTraits", []):
                if isinstance(trait, BinEmbedded):
                    champ_traits.extend(field.value for field in trait.fields if field.name.h == 0x053A1F33)
                else:
                    champ_traits.append(trait.h)

            spell_name = record.getv("spellNames")[0]
            spell_name = spell_name.rsplit("/", 1)[-1].lower()
            spell_key_name = None
            spell_key_tooltip = None
            for entry in tft_bin.entries:
                if entry.type == "SpellObject" and entry.getv("mScriptName").lower() == spell_name:
                    ability = entry.getv("mSpell")
                    ability_variables = [{"name": value.getv("mName"), "value": value.getv("mValues")} for value in ability.getv("mDataValues", [])]
                    if loc_keys := ability.get_path("mClientData", "mTooltipData", "mLocKeys"):
                        spell_key_name = loc_keys.get("keyName")
                        spell_key_tooltip = loc_keys.get("keyTooltip")
                    break
            else:
                ability_variables = []

            cost = record.getv("tier", None)
            if cost is None:
                rarity = champ.getv("mRarity", 0) + 1
                cost = rarity + int(rarity / 6)

            champs[name] = ({
                "apiName": champ.getv("mName"),
                "characterName": record.getv("mCharacterName"),
                "name": champ.getv(0xC3143D66),
                "cost": cost,
                "icon": champ.getv(0x466DC3CC) or champ.getv("mIconPath"),
                "tileIcon": champ.getv(0xDAC11DD4),
                "squareIcon": champ.getv(0x16071366),
                "traits": [traits[h]["name"] for h in champ_traits if h in traits],
                "stats": {
                    "hp": record.getv("baseHP"),
                    "mana": record["primaryAbilityResource"].value.getv("arBase", 100),
                    "initialMana": record.getv("mInitialMana", 0),
                    "damage": record.getv("BaseDamage"),
                    "armor": record.getv("baseArmor"),
                    "magicResist": record.getv("baseSpellBlock"),
                    "critMultiplier": record.getv("critDamageMultiplier"),
                    "critChance": record.getv("baseCritChance"),
                    "attackSpeed": record.getv("attackSpeed"),
                    "range": record.getv("attackRange", 0) // 180,
                },
                "ability": {
                    "name": champ.getv(0x87A69A5E) or spell_key_name,
                    "desc": champ.getv(0xBC4F18B3) or spell_key_tooltip,
                    "icon": champ.getv(0xDF0AD83B) or champ.getv("mPortraitIconPath"),
                    "variables": ability_variables,
                },
            }, champ_traits)

        return champs

    def parse_traits(self, map22):
        """Parse traits, return a map indexed by entry path"""
        trait_entries = [x for x in map22.entries if x.type == "TftTraitData"]

        traits = {}
        for trait in trait_entries:
            if "Template" in trait.getv("mName"):
                continue

            base_effect_list = []
            for trait_set in trait.getv(0x6f4cf34d, []):
                base_effect_list.extend(trait_set.getv("effectAmounts", []))

            effects = []
            if "mTraitSets" in trait:
                trait_sets = trait.getv("mTraitSets")
                field_prefix = 'm'
            else:
                trait_sets = trait.getv("mConditionalTraitSets", [])
                field_prefix = ''
            for trait_set in trait_sets:
                variables = {}
                for effect in base_effect_list + trait_set.getv("effectAmounts", []):
                    name = str(effect.getv("name")) if "name" in effect else "null"
                    variables[name] = effect.getv("value", "null")

                effects.append({
                    "minUnits": trait_set.getv(field_prefix + "MinUnits"),
                    "maxUnits": trait_set.getv(field_prefix + "MaxUnits") or 25000,
                    "style": trait_set.getv(field_prefix + "Style", 1),
                    "variables": variables,
                })

            traits[trait.path] = {
                "apiName": trait.getv("mName"),
                "name": trait.getv(0xC3143D66),
                "desc": trait.getv(0x765F18DA),
                "icon": trait.getv("mIconPath"),
                "effects": effects,
            }

        return traits


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="directory with extracted bin files")
    parser.add_argument("-o", "--output", default="tft", help="output directory")
    parser.add_argument('-V', '--patch-version', default=None,
                           help="patch version the input files belong to in the format XX.YY (default: latest patch)")
    args = parser.parse_args()

    parsed_version = PatchVersion(args.patch_version if args.patch_version else "main").as_int()

    tft_transformer = TftTransformer(args.input, game_version=parsed_version)
    tft_transformer.export(args.output, langs=None)

if __name__ == "__main__":
    main()



from .tools import BinaryParser


class SknFile:
    def __init__(self, file):
        if isinstance(file, str):
            file = open(file, "rb")
        if file.read(4) != b"\x33\x22\x11\x00":
            raise ValueError("missing magic code")

        f = BinaryParser(file)

        self.major, self.minor = f.unpack("<HH")

        if self.major == 0:
            index_count, vertex_count = f.unpack("<II")
            indices = [(f.unpack("<H")[0] + 1) for i in range(index_count)]
            vertices = [self.read_vertex(f) for i in range(vertex_count)]
            self.entries = [{"name": "Unknown", "vertices": vertices, "indices": indices}]
            return

        count, = f.unpack("<I")
        self.entries = [self.read_object(f) for i in range(count)]

        if self.major == 4:
            self.unknown, = f.unpack("<I")

        index_count, vertex_count = f.unpack("<II")

        if self.major == 4:
            self.vertex_size, = f.unpack("<I")
            self.vertex_type = f.unpack("<I")[0]
            self.bounding_box_min = f.unpack("<fff")
            self.bounding_box_max = f.unpack("<fff")
            self.bounding_sphere_location = f.unpack("<fff")
            self.bounding_sphere_radius, = f.unpack("<f")

        indices = [(f.unpack("<H")[0] + 1) for i in range(index_count)]
        vertices = [self.read_vertex(f) for i in range(vertex_count)]

        for entry in self.entries:
            entry["vertices"] = vertices[entry["start_vertex"] : entry["start_vertex"] + entry["vertex_count"]]
            entry["indices"] = [
                x - (0 if x < entry["start_vertex"] else entry["start_vertex"])
                for x in indices[entry["start_index"] : entry["start_index"] + entry["index_count"]]
            ]

            # remove redundant information
            del entry["start_vertex"]
            del entry["start_index"]
            del entry["vertex_count"]
            del entry["index_count"]

    def read_object(self, f):
        return {
            "name": f.unpack("64s")[0].split(b"\0", 1)[0].decode("utf-8"),
            "start_vertex": f.unpack("<I")[0],
            "vertex_count": f.unpack("<I")[0],
            "start_index": f.unpack("<I")[0],
            "index_count": f.unpack("<I")[0],
        }

    def read_vertex(self, f):
        return {
            "position": f.unpack("<fff"),
            "bone_indices": f.unpack("<BBBB"),
            "weight": f.unpack("<ffff"),
            "normal": f.unpack("<fff"),
            "uv": f.unpack("<ff"),
            "color": f.unpack("<BBBB") if hasattr(self, "vertex_type") and self.vertex_type >= 1 else None,
            "tangent": f.unpack("<ffff") if hasattr(self, "vertex_type") and self.vertex_type == 2 else None
        }

    def to_obj(self, entry) -> str:
        content = ""
        for vert in entry["vertices"]:
            content += "v %s %s %s\n" % vert["position"]
            content += "vt %s %s\n" % vert["uv"]
            content += "vn %s %s %s\n" % vert["normal"]

        for i in range(0, len(entry["indices"]), 3):
            a, b, c = entry["indices"][i : i + 3]
            content += "f {0}/{0}/{0} {1}/{1}/{1}/ {2}/{2}/{2}\n".format(a, b, c)

        return content



from enum import Enum


REGIONS = ['br', 'cn', 'eun', 'eune', 'euw', 'garena2', 'garena3', 'id', 'jp', 'kr', 'la', 'la1', 'la2', 'lan', 'las',
           'na', 'oc', 'oc1', 'oce', 'pbe', 'ph', 'ph2', 'ru', 'sg', 'tencent', 'th', 'th2', 'tr', 'tw', 'tw2', 'vn', 'vn2']

class Language(Enum):
    ar_ae = 'ar_ae'
    ar_eg = 'ar_eg'
    cs_cz = 'cs_cz'
    de_de = 'de_de'
    el_gr = 'el_gr'
    en_au = 'en_au'
    en_gb = 'en_gb'
    en_ph = 'en_ph'
    en_pl = 'en_pl'
    en_sg = 'en_sg'
    en_us = 'en_us'
    es_ar = 'es_ar'
    es_es = 'es_es'
    es_mx = 'es_mx'
    fr_fr = 'fr_fr'
    hu_hu = 'hu_hu'
    id_id = 'id_id'
    it_it = 'it_it'
    ja_jp = 'ja_jp'
    ko_kr = 'ko_kr'
    ms_my = 'ms_my'
    pl_pl = 'pl_pl'
    pt_br = 'pt_br'
    ro_ro = 'ro_ro'
    ru_ru = 'ru_ru'
    th_th = 'th_th'
    tr_tr = 'tr_tr'
    vi_vn = 'vi_vn'
    vn_vn = 'vn_vn'
    zh_cn = 'zh_cn'
    zh_my = 'zh_my'
    zh_tw = 'zh_tw'



import glob
import os
import re
import shutil
import struct
from contextlib import contextmanager

import pyzstd
zstd_decompress = pyzstd.decompress
try:
    import ujson
    def json_dump(obj, fp, **kwargs):
        return ujson.dump(obj, fp, **kwargs, escape_forward_slashes=False)
    def json_dumps(obj, **kwargs):
        return ujson.dumps(obj, **kwargs, escape_forward_slashes=False)
except ImportError:
    import json
    json_dump = json.dump
    json_dumps = json.dumps


@contextmanager
def write_file_or_remove(path, binary=True):
    """Open a file for writing, create its parent directory if needed

    If the writing fails, the file is removed.
    """
    try:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb' if binary else 'w') as f:
            yield f
    except:
        # remove partially written file
        try:
            os.remove(path)
        except OSError:
            pass
        raise

@contextmanager
def write_dir_or_remove(path):
    """Create a directory for writing, and its parent directory if needed

    If the writing fails, the directory and its content is removed.
    """
    try:
        os.makedirs(path, exist_ok=True)
        yield
    except:
        # remove directory
        shutil.rmtree(path, ignore_errors=True)
        raise


class BinaryParser:
    """Helper class to read from binary file object"""

    def __init__(self, f):
        self.f = f

    def tell(self):
        return self.f.tell()

    def seek(self, position):
        self.f.seek(position, 0)

    def skip(self, amount):
        self.f.seek(amount, 1)

    def rewind(self, amount):
        self.f.seek(-amount, 1)

    def unpack(self, fmt):
        length = struct.calcsize(fmt)
        return struct.unpack(fmt, self.f.read(length))

    def raw(self, length):
        return self.f.read(length)

    def unpack_string(self):
        """Unpack string prefixed by its 32-bit length"""
        return self.f.read(self.unpack('<L')[0]).decode('utf-8')

def stringtable_paths(base_dir, game):
    """
    Collect "stringtables" paths, indexed by locale (e.g. `en_us`)
    """

    # Path format history (from oldest to newest)
    # - data/menu/font_config_<lang>.txt
    # - data/menu/main_<lang>.stringtable
    # - <lang>/data/menu/en_us/main.stringtable
    #   (Directory is always `en_us`, cdtb has been adjusted to export to a language-specific subdir.)
    # - <lang>/data/menu/en_us/<game>.stringtable
    #   (game is either `lol` or `tft`)

    # Find the current format in order from newest to oldest
    lang_dict = {re.search(rf"(.._..)/data/menu/en_us/{game}\.stringtable$", path.replace('\\', '/')).group(1): path for path in glob.glob(os.path.join(base_dir, f"??_??/data/menu/en_us/{game}.stringtable"))}
    if not lang_dict:
        lang_dict = {re.search(r"(.._..)/data/menu/en_us/main\.stringtable$", path.replace('\\', '/')).group(1): path for path in glob.glob(os.path.join(base_dir, "??_??/data/menu/en_us/main.stringtable"))}
    if not lang_dict:
        lang_dict = {re.search(r"data/menu/main_(.._..)\.stringtable$", path.replace('\\', '/')).group(1): path for path in glob.glob(os.path.join(base_dir, "data/menu/main_??_??.stringtable"))}
    if not lang_dict:
        lang_dict = {re.search(r"data/menu/fontconfig_(.._..)\.txt$", path.replace('\\', '/')).group(1): path for path in glob.glob(os.path.join(base_dir, "data/menu/fontconfig_??_??.txt"))}
    if lang_dict:
        return lang_dict
    else:
        raise RuntimeError("cannot find stringtable files")

def convert_cdragon_path(path):
    path, ext = os.path.splitext(path.lower())
    if ext == ".dds" or ext == ".tex":
        ext = ".png"
    return path + ext



from xxhash import xxh3_64_intdigest, xxh64_intdigest
from base64 import b64encode
from .tools import BinaryParser
from .hashes import HashFile, default_hash_dir


def get_hashfile(game_version=1415):
    if game_version >= 1415:
        return hashfile_rst_xxh3
    else:
        return hashfile_rst_xxh64

def key_to_hash(key, bits=64, rsthash_version=1415):
    if isinstance(key, str):
        if rsthash_version >= 1415:
            key = xxh3_64_intdigest(key.lower())
        else:
            key = xxh64_intdigest(key.lower())
    return key & ((1 << bits) - 1)


hashfile_rst_xxh64 = HashFile(default_hash_dir / "hashes.rst.xxh64.txt", hash_size=16)
hashfile_rst_xxh3 = HashFile(default_hash_dir / "hashes.rst.xxh3.txt", hash_size=16)

class RstFile:
    def __init__(self, path_or_f=None, game_version=1415):
        self.font_config = None
        self.entries = {}
        self.hash_bits = 40
        self.version = None
        self.rsthash_version = game_version

        if path_or_f is not None:
            if isinstance(path_or_f, str):
                with open(path_or_f, "rb") as f:
                    self.parse_rst(f)
            else:
                self.parse_rst(path_or_f)

    def __getitem__(self, key):
        try:
            h = key_to_hash(key, self.hash_bits, self.rsthash_version)
            return self.entries[h]
        except (TypeError, KeyError):
            raise KeyError(key)

    def __contains__(self, key):
        try:
            h = key_to_hash(key, self.hash_bits, self.rsthash_version)
            return h in self.entries
        except TypeError:
            return False

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def parse_rst(self, f):
        parser = BinaryParser(f)

        magic, version = parser.unpack("<3sB")
        if magic != b'RST':
            raise ValueError("invalid magic code")

        if version == 2:
            if parser.unpack("<B")[0]:
                n, = parser.unpack("<L")
                self.font_config = parser.raw(n).decode("utf-8")
            else:
                self.font_config = None
        elif version == 3:
            pass
        elif version in (4, 5):
            self.hash_bits = 39
        else:
            raise ValueError(f"unsupported RST version: {version}")
        self.version = version

        hash_mask = (1 << self.hash_bits) - 1
        count, = parser.unpack("<L")
        entries = []
        for _ in range(count):
            v, = parser.unpack("<Q")
            entries.append((v >> self.hash_bits, v & hash_mask))

        has_trenc = False
        if version < 5:
            has_trenc = parser.unpack("<B")[0]

        data = parser.f.read()

        # Files are sometimes messed-up (e.g. windows-1252 quote)
        # Don't fail on UTF-8 decoding errors
        for i, h in entries:
            if has_trenc and data[i] == 0xFF:
                size = int.from_bytes(data[i+1:][:2], 'little')
                d = b64encode(data[i+3:][:size])
                self.entries[h] = d.decode('utf-8', 'replace')
            else:
                end = data.find(b"\0", i)
                d = data[i:end]
                self.entries[h] = d.decode('utf-8', 'replace')



#!/usr/bin/env python3
import os
import sys
import argparse
import textwrap
import fnmatch
import logging
from pathlib import Path
import cdtb
from cdtb.storage import (
    Storage,
    Patch,
    PatchVersion,
    parse_storage_component,
    storage_conf_from_path,
)
from cdtb.patcher import PatcherStorage
from cdtb.wad import Wad
from cdtb.export import CdragonRawPatchExporter
from cdtb.binfile import BinFile
from cdtb.sknfile import SknFile
from cdtb.hashes import (
    HashFile,
    LcuHashGuesser,
    GameHashGuesser,
    default_hashfile,
    default_hash_dir,
    update_default_hashfile,
)
from cdtb.tools import json_dump


def parse_component_arg(parser, storage: Storage, component: str):
    """Wrapper around parse_storage_component() to parse CLI arguments into patch elements"""
    try:
        component = parse_storage_component(storage, component)
    except ValueError:
        parser.error(f"invalid component: {component}")
    if component is None:
        parser.error(f"component not found: {component}")

    if isinstance(component, Patch):
        return list(component.elements)
    else:
        return [component]

def parse_component_args(parser, storage: Storage, components):
    return [e for c in components for e in parse_component_arg(parser, storage, c)]


def parse_storage_args(parser, args) -> Storage:
    """Parse storage-related arguments into a Storage"""

    default_path = os.environ.get('CDTB_STORAGE')

    path = default_path if args.storage is None else args.storage
    if path is None:
        conf = {
            'type': 'patcher',
            'path': 'cdn',
        }
    else:
        conf = storage_conf_from_path(path)
        if conf is None:
            parser.error(f"cannot retrieve storage configuration from '{path}'")
        if args.patchline is not None:
            if conf['type'] == 'patcher':
                conf['patchline'] = args.patchline
            else:
                parser.error("--patchline is only supported for 'patcher' storage")
    return Storage.from_conf(conf)


def command_download(parser, args):
    for component in parse_component_args(parser, args.storage, args.component):
        component.download(langs=args.langs)


def command_files(parser, args):
    for elem in parse_component_arg(parser, args.storage, args.component):
        it = elem.relpaths(langs=args.langs) if args.relative else elem.fspaths(langs=args.langs)
        for path in it:
            print(path)


def command_versions(parser, args):
    for patch in args.storage.patches(stored=args.stored):
        if args.type == 'patch' or args.type in (e.name for e in patch.elements):
            print(patch.version)


def command_fetch_hashes(parser, args):
    if default_hash_dir == Path(__file__).parent:
        if os.name == 'nt':
            user_dir = Path(os.environ.get('LOCALAPPDATA', 'LOCALAPPDATA'))
        else:
            user_dir = Path.home() / ".local/share"
        user_dir = user_dir / 'cdragon'
        parser.error(f"Cannot update hashes bundled with source; create {user_dir} to store them locally")

    print(f"Updating hash files: {default_hash_dir}")
    default_hash_dir.mkdir(parents=True, exist_ok=True)
    hash_files = [
        'hashes.binentries.txt',
        'hashes.binfields.txt',
        'hashes.binhashes.txt',
        'hashes.bintypes.txt',
        'hashes.game.txt',
        'hashes.lcu.txt',
        'hashes.rst.txt',
    ]
    for basename in hash_files:
        update_default_hashfile(basename)


def command_wad_extract(parser, args):
    if not os.path.isfile(args.wad):
        parser.error("WAD file does not exist")
    if not args.output:
        args.output = os.path.splitext(args.wad)[0]
    if os.path.exists(args.output) and not os.path.isdir(args.output):
        parser.error("output is not a directory")

    if args.hashes is None:
        hashfile = default_hashfile(args.wad)
    else:
        hashfile = HashFile(args.hashes)
    wad = Wad(args.wad, hashes=hashfile.load())
    if args.unknown == 'yes':
        pass  # don't filter
    elif args.unknown == 'only':
        wad.files = [wf for wf in wad.files if wf.path is None]
    elif args.unknown == 'no':
        wad.files = [wf for wf in wad.files if wf.path is not None]

    if args.pattern:
        wad.files = [wf for wf in wad.files if any(wf.path is not None and fnmatch.fnmatchcase(wf.path, p) for p in args.pattern)]

    wad.guess_extensions()
    wad.extract(args.output, overwrite=not args.lazy)


def command_wad_list(parser, args):
    if not os.path.isfile(args.wad):
        parser.error("WAD file does not exist")

    if args.hashes is None:
        hashfile = default_hashfile(args.wad)
    else:
        hashfile = HashFile(args.hashes)
    wad = Wad(args.wad, hashes=hashfile.load())

    wadfiles = [(wf.path or ('?.%s' % wf.ext if wf.ext else '?'), wf.path_hash) for wf in wad.files]
    for path, h in sorted(wadfiles):
        print(f"{h:016x} {path}")


def command_hashes_guess(parser, args):
    all_methods = [
        ("grep", "search for hashes in WAD files"),
        ("numbers", "substitute numbers in basenames"),
        ("basenames", "substitute basenames"),
        ("words", "substitute known words in basenames"),
        ("ext", "substitute extensions"),
        ("regionlang", "substitute region and lang (LCU only)"),
        ("plugin", "substitute plugin name (LCU only)"),
        ("skin-num", "substitute skinNN numbers (game only)"),
        ("character", "substitute character name (game only)"),
        ("prefixes", "check basename prefixes (game only)"),
    ]
    all_method_names = [name for name, _ in all_methods]

    if args.list_methods:
        name_width = max(len(name) for name in all_method_names)
        for name, desc in all_methods:
            print(f"  {name:{name_width}}  {desc}")
        return
    elif not args.wad:
        parser.error("neither \"wad\" nor \"--list-methods\" argument was found")

    if not args.methods:
        method_names = [name for name in all_method_names if name not in ("basenames", "words")]
    else:
        method_names = [s.strip() for s in args.methods.split(',')]
        for name in method_names:
            if name not in all_method_names:
                parser.error(f"unknown guessing method: {name}")

    # collect WAD paths
    wads = []
    for path_or_component in args.wad:
        try:
            component = parse_storage_component(args.storage, path_or_component)
        except ValueError:
            wads.append(Wad(path_or_component))
            continue
        if component is None:
            continue
        if isinstance(component, Patch):
            elements = component.elements
        else:
            elements = [component.elements]
        for elem in elements:
            wads.extend(Wad(p) for p in elem.fspaths() if p.endswith('.wad') or p.endswith('.wad.client'))

    # guess LCU hashes
    guesser = LcuHashGuesser.from_wads(wads)
    if guesser.unknown:
        nunknown = len(guesser.unknown)
        if "grep" in method_names:
            for wad in guesser.wads:
                wad.guess_extensions()
                guesser.grep_wad(wad)
        if "numbers" in method_names:
            guesser.substitute_numbers()
        if "basenames" in method_names:
            guesser.substitute_basenames()
        if "words" in method_names:
            guesser.substitute_basename_words()
        if "ext" in method_names:
            guesser.substitute_extensions()
        if "regionlang" in method_names:
            guesser.substitute_region_lang()
        if "plugin" in method_names:
            guesser.substitute_plugin()

        nfound = nunknown - len(guesser.unknown)
        if nfound:
            print(f"found LCU hashes: {nfound}")
            if not args.dry_run:
                guesser.save()

    # guess game hashes
    guesser = GameHashGuesser.from_wads(wads)
    if guesser.unknown:
        nunknown = len(guesser.unknown)
        if "grep" in method_names:
            for wad in guesser.wads:
                wad.guess_extensions()
                guesser.grep_wad(wad)
        if "numbers" in method_names:
            guesser.substitute_numbers()
        if "basenames" in method_names:
            guesser.substitute_basenames()
        if "words" in method_names:
            guesser.substitute_basename_words()
        if "ext" in method_names:
            guesser.substitute_extensions()
        if "skin-num" in method_names:
            guesser.substitute_skin_numbers()
        if "character" in method_names:
            guesser.substitute_character()
        if "prefixes" in method_names:
            guesser.check_basename_prefixes()

        nfound = nunknown - len(guesser.unknown)
        if nfound:
            print(f"found game hashes: {nfound}")
            if not args.dry_run:
                guesser.save()


def command_export(parser, args):
    storage = args.storage
    overwrite = not args.lazy
    symlinks = bool(args.symlinks)

    if not args.patch:
        # multiple patches (update only)
        if args.previous:
            parser.error("patch version required with --previous or --full")
        if args.first:
            parser.error("--from is required when no patch is provided")
        exporters = CdragonRawPatchExporter.from_directory(storage, args.output, PatchVersion(args.first), symlinks=symlinks)
        for exporter in exporters:
            exporter.process(overwrite=overwrite)
    else:
        if args.first:
            parser.error("--from cannot be used when providing a patch")
        # single patch
        # retrieve target and previous patch versions
        patch = storage.patch(None if args.patch == 'latest' else args.patch)
        if patch is None:
            parser.error(f"patch not found: {args.patch}")
        if args.previous == 'none':
            previous_patch = None
        elif args.previous:
            previous_patch = storage.patch(args.previous, stored=True)
            if previous_patch is None:
                parser.error(f"previous patch not found: {patch.version}")
        else:
            it = storage.patches(stored=True)
            for v in it:
                if v.version == patch.version:
                    previous_patch = next(it)
                    break
            else:
                parser.error("cannot guess previous patch")

        exporter = CdragonRawPatchExporter(os.path.join(args.output, str(patch.version)), patch, previous_patch, symlinks=symlinks)
        exporter.process(overwrite=overwrite)


def command_skn_extract(parser, args):
    if not os.path.isfile(args.skn):
        parser.error(f"SKN file not found: {args.skn}")

    sknfile = SknFile(args.skn)
    if args.output is None:
        args.output = os.path.splitext(args.skn)[0]
    os.makedirs(args.output, exist_ok=True)
    for entry in sknfile.entries:
        name = os.path.join(args.output, entry["name"] + ".obj")
        with open(name, "w") as f:
            f.write(sknfile.to_obj(entry))


def command_bin_dump(parser, args):
    if not os.path.isfile(args.bin):
        parser.error(f"BIN file not found: {args.bin}")

    parsed_version = PatchVersion(args.patch_version if args.patch_version else "main").as_int()

    with open(args.bin, 'rb') as f:
        binfile = BinFile(f, btype_version=parsed_version)
    if args.json:
        json_dump(binfile.to_serializable(), sys.stdout)
    else:
        for entry in binfile.entries:
            print(entry)
        if binfile.patch_entries is not None:
            for entry in binfile.patch_entries:
                print(entry)


def create_parser():
    parser = argparse.ArgumentParser('cdtb',
        description="Toolbox to work with League of Legends game and client files",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent(f"""
            The following formats are supported for components:

              patch=version    patch with given version
              game=version     game files for patch with given version
              client=version   client files (LCU) for patch with given version

            The following formats are supported for patch version:

              X.Y       patch X.Y (all subpatches)
              X.Y.      latest subpatch for patch X.Y (latest elements)
              <empty>   latest available subpatch

            Hashes directory: {default_hash_dir}

            Environment variables

              CDTB_STORAGE     default `--storage` value
              CDTB_EXPORT      default 'export --output` value
              CDTB_HASHES_DIR  path to directory with hash files
              CDRAGON_DATA               path to `Data` repository, for hash files

        """),
    )

    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="be verbose")

    subparsers = parser.add_subparsers(dest='command', help="command")

    default_export = os.environ.get('CDTB_EXPORT', 'export')

    # storage arguments
    storage_parser = argparse.ArgumentParser(add_help=False)
    storage_parser.add_argument('-s', '--storage', default=None,
                                help="path to downloaded files, with an optional storage type prefix (`type:path`)")
    storage_parser.add_argument('--patchline', choices=["pbe", "live"], default=None,
                                help="select a patchline")

    # component-based commands

    component_parser = argparse.ArgumentParser(add_help=False, parents=[storage_parser])
    component_parser.add_argument('--no-lang', dest='langs', action='store_false', default=True,
                                  help="ignore language projects from solutions")
    component_parser.add_argument('--lang', dest='langs', nargs='*',
                                  help="use projects from solutions in given languages (default: all)")

    subparser = subparsers.add_parser('download', parents=[component_parser],
                                      help="download components to the storage")
    subparser.add_argument('component', nargs='+',
                           help="components to download")

    subparser = subparsers.add_parser('files', parents=[component_parser],
                                      help="list files of a component")
    subparser.add_argument('-r', '--relative', action='store_true',
                           help="print relative export path insted of filesystem path")
    subparser.add_argument('component',
                           help="project version, solution version or patch version")

    subparser = subparsers.add_parser('versions', parents=[storage_parser],
                                      help="list available patch versions for a component")
    subparser.add_argument('-a', '--all', dest='stored', action='store_false', default=True,
                           help="when listing patch versions, don't use only stored solutions")
    subparser.add_argument('type', choices={'patch', 'game', 'client'},
                           help="display versions of given component type")


    # Tooling and maintenance

    subparser = subparsers.add_parser('fetch-hashes',
                                      help="download up-to-date hash lists")


    # WAD commands

    subparser = subparsers.add_parser('wad-extract',
                                      help="extract a WAD file")
    subparser.add_argument('-o', '--output',
                           help="extract directory")
    subparser.add_argument('-H', '--hashes',
                           help="hashes of known paths")
    subparser.add_argument('-p', '--pattern', action='append',
                           help="extract only files matching pattern with shell-like wildcards")
    subparser.add_argument('-u', '--unknown', choices=('yes', 'only', 'no'), default='yes',
                           help="control extract of unknown files (default: %(default)s)")
    subparser.add_argument('--lazy', action='store_true',
                           help="don't overwrite files, assume they are already correctly extracted")
    subparser.add_argument('wad',
                           help="WAD file to extract")

    subparser = subparsers.add_parser('wad-list',
                                      help="list WAD content")
    subparser.add_argument('-H', '--hashes',
                           help="hashes of known paths")
    subparser.add_argument('wad',
                           help="WAD file to list")

    subparser = subparsers.add_parser('hashes-guess', parents=[storage_parser],
                                      help="guess hashes from WAD content")
    subparser.add_argument('-n', '--dry-run', action='store_true',
                           help="list new hashes but don't update the hashes file")
    subparser.add_argument('-m', '--methods',
                           help="list of guessing methods to run, comma-separated (default: all except \"basenames\" and \"words\")")
    subparser.add_argument('--list-methods', action='store_true',
                           help="display a list of valid guessing methods and exit")
    subparser.add_argument('wad', nargs='*',
                           help="WAD files or components to analyze")


    # export command

    subparser = subparsers.add_parser('export', parents=[storage_parser],
                                      help="export files to directories, separated by patch")
    subparser.add_argument('-o', '--output', default=default_export,
                           help="directory for files to export (default: %(default)s)")
    subparser.add_argument('-L', '--symlinks', action='store_true',
                           help="create symlinks (if supported by platform)")
    subparser.add_argument('--previous',
                           help="previous patch version to compare with (default: guessed)")
    subparser.add_argument('--full', dest='previous', action='store_const', const='none',
                           help="export the whole patch (don't compare with a previous one)")
    subparser.add_argument('--from', dest='first',
                           help="if a patch is not provided, update all exported patches starting from this one")
    subparser.add_argument('--lazy', action='store_true',
                           help="don't overwrite files, assume they are already correctly extracted")
    subparser.add_argument('patch', nargs='?',
                           help="patch version to export or 'latest', can be omitted to update all exported patches")


    # bin files commands

    subparser = subparsers.add_parser('bin-dump',
                                      help="dump a BIN file as a text tree")
    subparser.add_argument('-j', '--json', action='store_true',
                           help="extract to JSON")
    subparser.add_argument('-V', '--patch-version', default=None,
                           help="patch version this BIN file belongs to in the format XX.YY (default: latest patch)")
    subparser.add_argument('bin',
                           help="BIN file to extract")

    # skn files commands

    subparser = subparsers.add_parser('skn-extract',
                                      help="extract an SKN file to a directory")
    subparser.add_argument('-o', '--output',
                           help="output directory")
    subparser.add_argument('skn',
                           help="SKN file to extract")

    return parser


def main(argv = None):
    parser = create_parser()
    args = parser.parse_args(argv)

    if not args.command:
        parser.print_help()
        return

    if args.verbose >= 3:
        loglevel = logging.DEBUG
    elif args.verbose >= 1:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING

    logging.basicConfig(
        level=loglevel,
        datefmt='%H:%M:%S',
        format='%(asctime)s %(levelname)s %(name)s - %(message)s',
    )

    logger = cdtb.logger
    if args.verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif args.verbose >= 1:
        logger.setLevel(logging.INFO)

    if hasattr(args, 'storage'):
        args.storage = parse_storage_args(parser, args)

    globals()[f"command_{args.command.replace('-', '_')}"](parser, args)


if __name__ == "__main__":
    main()



import os
import re
import io
import json
import time
import hashlib
import logging
from typing import List, Optional, Generator

from .storage import (
    Storage,
    PatchElement,
    PatchVersion,
    get_system_yaml_version,
    get_content_metadata_version,
)
from .tools import (
    BinaryParser,
    write_file_or_remove,
    zstd_decompress,
)

logger = logging.getLogger(__name__)


class PatcherChunk:
    def __init__(self, chunk_id, bundle, offset, size, target_size):
        self.chunk_id = chunk_id
        self.bundle = bundle
        self.offset = offset
        self.size = size
        self.target_size = target_size

class PatcherBundle:
    def __init__(self, bundle_id):
        self.bundle_id = bundle_id
        self.chunks = []

    def add_chunk(self, chunk_id, size, target_size):
        try:
            last_chunk = self.chunks[-1]
            offset = last_chunk.offset + last_chunk.size
        except IndexError:
            offset = 0
        self.chunks.append(PatcherChunk(chunk_id, self, offset, size, target_size))

class PatcherFile:
    def __init__(self, name, size, link, flags, chunks):
        self.name = name
        self.size = size
        self.link = link
        self.flags = flags
        self.chunks = chunks

    def hexdigest(self):
        """Compute a hash unique for this file content"""
        m = hashlib.sha1()
        for chunk in self.chunks:
            m.update(b"%016X" % chunk.chunk_id)
        return m.hexdigest()

    @staticmethod
    def langs_predicate(langs):
        """Return a predicate function for a locale filtering parameter"""
        if langs is False:
            # assume only locales flags follow this pattern
            return lambda f: f.flags is None or not any('_' in f and len(f) == 5 for f in f.flags)
        elif langs is True:
            return lambda f: True
        else:
            lang = langs.lower()  # compare lowercased
            return lambda f: f.flags is not None and any(f.lower() == lang for f in f.flags)


class PatcherManifest:
    def __init__(self, path_or_f=None):
        self.bundles = None
        self.chunks = None
        self.flags = None
        self.files = None

        if path_or_f is not None:
            if isinstance(path_or_f, str):
                with open(path_or_f, "rb") as f:
                    self.parse_rman(f)
            else:
                self.parse_rman(f)

    def filter_files(self, langs=True) -> List[PatcherFile]:
        """Filter files from the manifest with provided filters"""
        return filter(PatcherFile.langs_predicate(langs), self.files.values())

    def parse_rman(self, f):
        parser = BinaryParser(f)

        magic, version_major, version_minor = parser.unpack("<4sBB")
        if magic != b'RMAN':
            raise ValueError("invalid magic code")
        if (version_major, version_minor) != (2, 0):
            raise ValueError(f"unsupported RMAN version: {version_major}.{version_minor}")

        flags, offset, length, _manifest_id, _body_length = parser.unpack("<HLLQL")
        assert flags & (1 << 9)  # other flags not handled
        assert offset == parser.tell()

        f = io.BytesIO(zstd_decompress(parser.raw(length)))
        return self.parse_body(f)

    def parse_body(self, f):
        parser = BinaryParser(f)

        # header (unknown values, skip it)
        n, = parser.unpack('<l')
        parser.skip(n)

        # offsets to tables (convert to absolute)
        offsets_base = parser.tell()
        offsets = list(offsets_base + 4*i + v for i, v in enumerate(parser.unpack('<6l')))

        parser.seek(offsets[0])
        self.bundles = list(self._parse_table(parser, self._parse_bundle))

        parser.seek(offsets[1])
        self.flags = dict(self._parse_table(parser, self._parse_flag))

        # build a list of chunks, indexed by ID
        self.chunks = {chunk.chunk_id: chunk for bundle in self.bundles for chunk in bundle.chunks}

        parser.seek(offsets[2])
        file_entries = list(self._parse_table(parser, self._parse_file_entry))
        parser.seek(offsets[3])
        directories = {did: (name, parent) for name, did, parent in self._parse_table(parser, self._parse_directory)}

        # merge files and directory data
        self.files = {}
        for name, link, flag_ids, dir_id, filesize, chunk_ids in file_entries:
            while dir_id is not None:
                dir_name, dir_id = directories[dir_id]
                name = f"{dir_name}/{name}"
            if flag_ids is not None:
                flags = [self.flags[i] for i in flag_ids]
            else:
                flags = None
            file_chunks = [self.chunks[chunk_id] for chunk_id in chunk_ids]
            self.files[name] = PatcherFile(name, filesize, link, flags, file_chunks)

        # note: last two tables are unresolved


    @staticmethod
    def _parse_table(parser, entry_parser):
        count, = parser.unpack('<l')

        for _ in range(count):
            pos = parser.tell()
            offset, = parser.unpack('<l')
            parser.seek(pos + offset)
            yield entry_parser(parser)
            parser.seek(pos + 4)

    @classmethod
    def _parse_bundle(cls, parser):
        """Parse a bundle entry"""

        def parse_chunklist(parser):
            fields = cls._parse_field_table(parser, (
                ('chunk_id', '<Q'),
                ('compressed_size', '<L'),
                ('uncompressed_size', '<L'),
            ))
            return fields['chunk_id'], fields['compressed_size'], fields['uncompressed_size']

        fields = cls._parse_field_table(parser, (
            ('bundle_id', '<Q'),
            ('chunks_offset', 'offset'),
        ))

        bundle = PatcherBundle(fields['bundle_id'])
        parser.seek(fields['chunks_offset'])
        for (chunk_id, compressed_size, uncompressed_size) in cls._parse_table(parser, parse_chunklist):
            bundle.add_chunk(chunk_id, compressed_size, uncompressed_size)

        return bundle

    @staticmethod
    def _parse_flag(parser):
        parser.skip(4)  # skip offset table offset
        flag_id, offset, = parser.unpack('<xxxBl')
        parser.skip(offset - 4)
        return (flag_id, parser.unpack_string())

    @classmethod
    def _parse_file_entry(cls, parser):
        """Parse a file entry
        (name, link, flag_ids, directory_id, filesize, chunk_ids)
        """
        fields = cls._parse_field_table(parser, (
            ('file_id', '<Q'),
            ('directory_id', '<Q'),
            ('file_size', '<L'),
            ('name', 'str'),
            ('flags', '<Q'),
            None,
            None,
            ('chunks', 'offset'),
            None,
            ('link', 'str'),
            None,
            None,
            None,
        ))

        flag_mask = fields['flags']
        if flag_mask:
            flag_ids = [i+1 for i in range(64) if flag_mask & (1 << i)]
        else:
            flag_ids = None

        parser.seek(fields['chunks'])
        chunk_count, = parser.unpack('<L')  # _ == 0
        chunk_ids = list(parser.unpack(f'<{chunk_count}Q'))

        return (fields['name'], fields['link'], flag_ids, fields['directory_id'], fields['file_size'], chunk_ids)

    @classmethod
    def _parse_directory(cls, parser):
        """Parse a directory entry
        (name, directory_id, parent_id)
        """
        fields = cls._parse_field_table(parser, (
            ('directory_id', '<Q'),
            ('parent_id', '<Q'),
            ('name', 'str'),
        ))
        return (fields['name'], fields['directory_id'], fields['parent_id'])

    @staticmethod
    def _parse_field_table(parser, fields):
        entry_pos = parser.tell()
        fields_pos = entry_pos - parser.unpack('<l')[0]
        nfields = len(fields)
        output = {}
        parser.seek(fields_pos)
        parser.skip(2) # vtable size
        parser.skip(2) # object size
        for _, field, offset in zip(range(nfields), fields, parser.unpack(f'<{nfields}H')):
            if field is None:
                continue
            name, fmt = field
            if offset == 0 or fmt is None:
                value = None
            else:
                pos = entry_pos + offset
                parser.seek(pos)
                if fmt == 'offset':
                    value = pos + parser.unpack('<l')[0]
                elif fmt == 'str':
                    value = parser.unpack('<l')[0]
                    parser.seek(pos + value)
                    value = parser.unpack_string()
                else:
                    value = parser.unpack(fmt)[0]
            output[name] = value
        return output


class PatcherStorage(Storage):
    """
    Storage based on CDN with bundles and chunks

    File tree:
      channels/  -- mirror of CDN's channels/ directory
        public/
          bundles/
          releases/
      cdtb/  -- CDTB files and exported files
        channels/  -- obsolete (used for the previous Riot release versionning)
        releases/{patchline}/{timestamp}/  -- one directory per release (game/client pair)
          release.json  -- release information (notably manifest URLs)
          files/  -- release files, extracted from bundles
          patch_version.{elem}  -- cached patch version for element `name`
        releases/{patchline}/{region}/latest.timestamp  -- last timestamp version
        files/  -- extracted files, named after their hash (shared)

    One instance handles a single patchline, even if all channels are stored
    under the same file tree. This is because only bundles and chunks are
    shared between channels, not versions and extracted files.

    The `cdtb/files/` directory is used to avoid to extract (and store)
    multiple copies of the same file. Actual files are put in `cdtb/files/` and
    `channels/.../files/` contains symlinks to them.
    This can be disabled by setting the `use_extract_symlinks` option to false.

    Sometimes, HTTPS requests to clientconfig.rpg.riotgames.com are denied.
    If `clientconfig_data` is set, the provided file (if available) or URL is used.

    Configuration options:
      patchline -- the patchline name (`live` or `pbe`)
      region -- region from which use configuration
      use_extract_symlinks -- if false, disable use of symlinks for extracted files
      clientconfig_data -- file or URL to use as 'clientconfig.rpg.riotgames.com' data

    """

    storage_type = 'patcher'

    URL_BASE = "https://lol.dyn.riotcdn.net/"
    DEFAULT_PATCHLINE = 'live'
    CLIENT_LIVE_REGION = 'EUW'
    GAME_LIVE_PLATFORM = 'EUW1'

    def __init__(self, path, patchline=DEFAULT_PATCHLINE):
        super().__init__(path, self.URL_BASE)
        self.patchline = patchline
        self.use_extract_symlinks = True
        self.clientconfig_data = None

    @classmethod
    def from_conf_data(cls, conf):
        storage = cls(conf['path'], conf.get('patchline', cls.DEFAULT_PATCHLINE))
        if conf.get('use_extract_symlinks') is False:
            storage.use_extract_symlinks = False
        if 'clientconfig_data' in conf:
            storage.clientconfig_data = conf['clientconfig_data']
        return storage

    def base_release_path(self):
        return self.fspath(f"cdtb/releases/{self.patchline}")

    def iter_releases(self) -> List['PatcherRelease']:
        """Generate releases available in the storage, latest first"""

        base = self.base_release_path()
        if not os.path.isdir(base):
            return
        versions = []
        for version in os.listdir(base):
            try:
                versions.append(int(version))
            except ValueError:
                continue  # ignore extra files

        for version in sorted(versions, reverse=True):
            if os.path.isfile(f"{base}/{version}/release.json"):
                yield PatcherRelease(self, version)

    def patch_elements(self, stored=False):
        if not stored:
            # add latest release to the storage, if any
            self.fetch_latest_update()

        for release in self.iter_releases():
            for elem in release.elements():
                yield PatcherPatchElement(elem)

    def fetch_latest_update(self):
        """Fetch the latest release from the CDN"""

        release_info = {
            'client_patch_url': self.get_latest_client_manifest(),
            'game_patch_url': self.get_latest_game_manifest(),
        }

        latest_timestamp = self.latest_timestamp()
        if latest_timestamp is None:
            latest_release = None
        else:
            latest_release = PatcherRelease(self, latest_timestamp)
        if latest_release and release_info == latest_release.data:
            return  # already up to date

        # Create the release information file and update `latest.timestamp`
        timestamp = int(time.time())
        base = self.base_release_path()
        with write_file_or_remove(f"{base}/{timestamp}/release.json", binary=False) as f:
            json.dump(release_info, f)
        with write_file_or_remove(f"{base}/latest.timestamp", binary=False) as f:
            print(f"{timestamp}", file=f)

    def get_latest_client_manifest(self):
        url_or_path = self.clientconfig_data
        if url_or_path and not url_or_path.startswith('http://') and not url_or_path.startswith('https://') and os.path.exists(url_or_path):
            with open(url_or_path) as f:
                data = json.load(f)
        else:
            url = url_or_path or "https://clientconfig.rpg.riotgames.com/api/v1/config/public?namespace=keystone.products.league_of_legends.patchlines"
            r = self.s.get(url)
            r.raise_for_status()
            data = r.json()
        region = 'PBE' if self.patchline == 'pbe' else self.CLIENT_LIVE_REGION
        for config in data[f"keystone.products.league_of_legends.patchlines.{self.patchline}"]["platforms"]["win"]["configurations"]:
            if config['id'] == region:
                return config['patch_url']
        raise ValueError(f"client configuration not found for {self.patchline}")

    def get_latest_game_manifest(self):
        platform = 'PBE1' if self.patchline == 'pbe' else self.GAME_LIVE_PLATFORM
        r = self.s.get(f"https://sieve.services.riotcdn.net/api/v1/products/lol/version-sets/{platform}?q[platform]=windows&q[published]=true")
        r.raise_for_status()
        data = r.json()
        assert len(data["releases"]) == 1
        return data["releases"][-1]["download"]["url"]

    def latest_timestamp(self):
        path = f"{self.base_release_path()}/latest.timestamp"
        if not os.path.exists(path):
            return None
        with open(path) as f:
            return int(f.read().strip())

    def download_manifest(self, id_or_url):
        """Download a manifest from its ID or full URL if needed, return its path in the storage"""

        if isinstance(id_or_url, str):
            id_or_url = id_or_url.replace(".secure.", ".")
            if not id_or_url.startswith(self.url):
                raise ValueError(f"unexpected base URL for manifest: {id_or_url}")
            m = re.match(r"^channels/public/releases/([0-9A-F]{16})\.manifest$", id_or_url[len(self.url):])
            if not m:
                raise ValueError(f"unexpected manifest URL format: {id_or_url}")
            manif_id = int(m.group(1), 16)
        else:
            manif_id = id_or_url

        path = f"channels/public/releases/{manif_id:016X}.manifest"
        self.download(path, None)
        return path

    def download_bundle(self, bundle_id):
        """Download a bundle from its ID, return its path in the storage"""

        path = f"channels/public/bundles/{bundle_id:016X}.bundle"
        self.download(path, None)
        return path

    def load_chunk(self, chunk: PatcherChunk):
        """Load chunk data from a bundle"""
        path = f"channels/public/bundles/{chunk.bundle.bundle_id:016X}.bundle"
        with open(self.fspath(path), "rb") as f:
            f.seek(chunk.offset)
            # assume chunk is compressed
            return zstd_decompress(f.read(chunk.size))

    def extract_file(self, file: PatcherFile, output, overwrite=False):
        """Extract a file from its chunks, which must be available"""

        if not overwrite and os.path.isfile(output) and os.path.getsize(output) == file.size:
            logger.debug(f"skip {file.name}: already built to {output}")
            return

        if self.use_extract_symlinks:
            real_output = self.fspath(f"cdtb/files/{file.hexdigest()}")
        else:
            real_output = output

        if not os.path.isfile(real_output):
            logger.debug(f"extract {file.name} to {real_output}")
            with write_file_or_remove(real_output) as f:
                for chunk in file.chunks:
                    f.write(self.load_chunk(chunk))

        if self.use_extract_symlinks:
            logger.debug(f"symlink {real_output} to {output}")
            try:
                os.remove(output)
            except OSError:
                pass
            output_dir = os.path.dirname(output)
            os.makedirs(output_dir, exist_ok=True)
            os.symlink(os.path.relpath(real_output, output_dir), output)


class PatcherRelease:
    """
    A single released version
    """

    def __init__(self, storage: PatcherStorage, version):
        self.storage = storage
        self.version = version
        self.storage_dir = f"{storage.base_release_path()}/{version}"
        with open(f"{self.storage_dir}/release.json") as f:
            self.data = json.load(f)

    def __str__(self):
        return f"patcher:v{self.version}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.version}>"

    def element(self, name) -> Optional['PatcherReleaseElement']:
        """Retrieve element with given name, None if not available"""

        if not self.data.get(f"{name}_patch_url"):
            return None
        return PatcherReleaseElement(self, name)

    def available_elements(self):
        """Return the name of elements available in this release"""
        return [name for name in PatchElement.names if self.data.get(f"{name}_patch_url")]

    def elements(self) -> Generator['PatcherReleaseElement', None, None]:
        """Iterate over available elements"""
        for name in PatchElement.names:
            elem = self.element(name)
            if elem is not None:
                yield elem

    def download_bundles(self, langs=True):
        """Download bundles from CDN for the release"""

        for elem in self.elements():
            elem.download(langs=langs)

    def extract(self, langs=True, overwrite=False):
        """Extract release files from downloaded bundles"""

        for elem in self.elements():
            elem.extract(langs=langs, overwrite=overwrite)


class PatcherReleaseElement:
    """
    Element of a release (game or client)
    """

    def __init__(self, release: PatcherRelease, name):
        self.release = release
        self.name = name
        self._manif = None
        self.manif_url = self.release.data.get(f"{self.name}_patch_url")

    @property
    def manif(self):
        if self._manif is None:
            if not self.manif_url:
                raise ValueError("no manifest URL found for {self}")
            path = self.release.storage.download_manifest(self.manif_url)
            self._manif = PatcherManifest(self.release.storage.fspath(path))
        return self._manif

    def __str__(self):
        return f"patcher:v{self.release.version}:{self.name}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.release.version} {self.name}>"

    def bundle_ids(self, langs=True) -> set:
        """Return IDs of bundles used by the element as a set"""

        files = [f for f in self.manif.filter_files(langs) if not f.link]
        return {chunk.bundle.bundle_id for f in files for chunk in f.chunks}

    def download_bundles(self, langs=True):
        """Download bundles from CDN"""

        logger.info(f"download bundles for {self}")
        for bundle_id in sorted(self.bundle_ids(langs=langs)):
            self.release.storage.download_bundle(bundle_id)

    def extract(self, langs=True, overwrite=False):
        """Extract files to the storage"""

        logger.info(f"extract files from {self}")
        files = [f for f in self.manif.filter_files(langs) if not f.link]
        for file in sorted(files, key=lambda f: f.name):
            self.extract_file(file, overwrite=overwrite)

    def extract_path(self, file: PatcherFile):
        """Return the path to which a file is extracted"""
        return f"{self.release.storage_dir}/files/{file.name}"

    def extract_file(self, file: PatcherFile, overwrite=False):
        """Extract a single file"""
        self.release.storage.extract_file(file, self.extract_path(file), overwrite=overwrite)

    def patch_version(self) -> Optional[PatchVersion]:
        """Return patch version or None if there is none

        This method reads/writes version from/to cache.
        """

        # for PBE: version is always "main"
        if self.release.storage.patchline == "pbe":
            return PatchVersion("main")

        cache = f"{self.release.storage_dir}/patch_version.{self.name}"
        if os.path.isfile(cache):
            logger.debug(f"retrieving patch version for {self} from cache")
            with open(cache) as f:
                version = f.read().strip()
                version = PatchVersion(version) if version else None
        else:
            version = self._retrieve_patch_version()
            if version is None:
                logger.warning(f"failed to retrieve patch version of {self}")
            else:
                with open(cache, 'w') as f:
                    f.write(f"{version}\n")
        return version

    def _retrieve_patch_version(self) -> Optional[PatchVersion]:
        """Retrieve patch version from files (no cache handling)

        Return None if there is no patch version (for instance, because files
        are not available anymore).
        Raise an exception if patch version cannot be retrieved.
        """

        logger.debug(f"retrieving patch version for {self}")

        retrievers = {
            # elem_name: (file_name, extractor)
            'client': ('system.yaml', get_system_yaml_version),
            'game': ('content-metadata.json', get_content_metadata_version),
        }

        file_name, extractor = retrievers[self.name]
        file = self.manif.files[file_name]
        # download and extract file if needed
        for bundle_id in {chunk.bundle.bundle_id for chunk in file.chunks}:
            self.release.storage.download_bundle(bundle_id)
        self.extract_file(file)

        version = extractor(self.extract_path(file))
        return PatchVersion(version)


class PatcherPatchElement(PatchElement):
    """Patch element from a patcher storage"""

    def __init__(self, elem: PatcherReleaseElement):
        self.elem = elem
        version = elem.patch_version()
        super().__init__(elem.name, version)

    def download(self, langs=True):
        self.elem.download_bundles(langs=langs)
        self.elem.extract(langs=langs)

    def fspaths(self, langs=True):
        return (self.elem.extract_path(f) for f in self.elem.manif.filter_files(langs=langs))

    def relpaths(self, langs=True):
        return (f.name.lower() for f in self.elem.manif.filter_files(langs=langs))

    def paths(self, langs=True):
        for f in self.elem.manif.filter_files(langs=langs):
            yield (self.elem.extract_path(f), f.name.lower())




import os
import sys
import re
import datetime
import glob
import itertools
import signal
import time
import json
import struct
import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Dict
import requests
from xxhash import xxh64_intdigest
from .data import REGIONS, Language
from .tools import write_file_or_remove

logger = logging.getLogger(__name__)


def _default_hash_dir():
    """
    Hash directory is search for in this order

    - `$CDTB_HASHES_DIR` if set
    - `$CDRAGON_DATA/hashes/lol` if `$CDRAGON_DATA` is set
    - If one of the following "data" directory exists, use `<dir>/hashes/lol`
      - `$XDG_DATA_HOME/cdragon`
      - `$LOCALAPPDATA/cdragon`
      - `~/.local/share/cdragon` (see `Path.home()` for `~`)

    """
    def _env_dir(name):
        value = os.environ.get(name)
        return Path(value) if value else None
    if path := _env_dir('CDTB_HASHES_DIR'):
        return path
    if path := _env_dir('CDRAGON_DATA'):
        return path / 'hashes/lol'

    # Always search in XDG_DATA_HOME, even on Windows
    path = _env_dir('XDG_DATA_HOME')
    if not path or not path.is_absolute():
        # Assume LOCALAPPDATA is not set when not on Windows
        path = _env_dir('LOCALAPPDATA') or Path.home() / ".local/share"
    return path / 'cdragon/data/hashes/lol'

default_hash_dir = _default_hash_dir()


class HashFile:
    """Store hashes, support save/load and caching"""

    def __init__(self, filename, hash_size=16):
        self.filename = filename
        self.line_format = f"{{:0{hash_size}x}} {{}}"
        self.hashes = None

    def load(self, force=False) -> Dict[int, str]:
        if force or self.hashes is None:
            try:
                with open(self.filename) as f:
                    hashes = (l.split(' ', 1) for l in f)
                    self.hashes = {int(h, 16): s.rstrip('\n') for h, s in hashes}
            except FileNotFoundError:
                raise FileNotFoundError(f"Hash file not found; try to run 'fetch-hashes' command: {self.filename}")
        return self.hashes

    def save(self):
        with open(self.filename, 'w', newline='') as f:
            for h, s in sorted(self.hashes.items(), key=lambda kv: kv[1]):
                print(self.line_format.format(h, s), file=f)

hashfile_lcu = HashFile(default_hash_dir / "hashes.lcu.txt")
hashfile_game = HashFile(default_hash_dir / "hashes.game.txt")

def default_hashfile(path):
    """Return the default hashfile for the given WAD file"""
    if path.endswith(".wad.client"):
        return hashfile_game
    elif path.endswith(".wad"):
        return hashfile_lcu
    else:
        raise ValueError(f"no default hashes for WAD file '{path}'")

def update_default_hashfile(basename):
    """Update a hashfile if a new version is available for download"""

    path = default_hash_dir / basename
    url = f"https://raw.communitydragon.org/data/hashes/lol/{basename}"

    try:
        last_time = path.stat().st_mtime
        # CloudFlare does not support 'if-modified-since'
        # We have to send a separate HEAD
        r = requests.head(url)
        r.raise_for_status()
        last_modified = datetime.datetime.strptime(r.headers['last-modified'], '%a, %d %b %Y %H:%M:%S GMT').replace(tzinfo=datetime.timezone.utc)
        if last_time >= last_modified.timestamp():
            return  # up to date
    except FileNotFoundError:
        pass  # Never downloaded

    logger.debug(f"update hash file from {url}")
    r = requests.get(url)
    r.raise_for_status()
    with write_file_or_remove(path) as f:
        f.write(r.content)


def build_wordlist(paths):
    """Build a list of words from paths"""

    words = set()
    re_split = re.compile(r'[/_.-]')
    for path in paths:
        words |= set(re_split.split(path)[:-1])

    # filter out large numbers
    re_filter_words = re.compile(r'^[0-9]{3,}$')
    words = set(w for w in words if not re_filter_words.search(w))
    return sorted(words)


@contextmanager
def sigint_callback(callback):
    """Context to display progress of long operations on SIGINT"""

    tlast = 0
    def signal_handler(signal, frame):
        nonlocal tlast
        callback()
        tnow = time.time()
        if tnow - tlast < 0.5:
            raise KeyboardInterrupt()
        tlast = tnow

    previous_handler = signal.signal(signal.SIGINT, signal_handler)
    try:
        yield
    finally:
        _handler_back = signal.signal(signal.SIGINT, previous_handler)

def progress_iterate(sequence, formatter=None):
    """Iterate over sequence, display progress on SIGINT"""

    if formatter is None:
        formatter = lambda v: v

    interrupted = False
    def handler():
        nonlocal interrupted
        interrupted = True
    with sigint_callback(handler):
        for i, v in enumerate(sequence):
            if interrupted:
                print("  %5.1f%%  %s" % (100 * i / len(sequence), formatter(v)))
                interrupted = False
            yield v

def progress_iterator(sequence, formatter=None):
    # don't handle SIGINT if output is not a terminal
    if not os.isatty(sys.stderr.fileno()):
        return sequence
    return progress_iterate(sequence, formatter)



class HashGuesser:
    """
    Guess hashes from files
    """

    def __init__(self, hashfile, hashes):
        self.hashfile = hashfile
        if not isinstance(hashes, set):
            hashes = set(hashes)

        self.known = self.hashfile.load()
        self.unknown = hashes - set(self.known)
        self.wads = None
        self.__directory_list = None  # cache

    @classmethod
    def from_wads(cls, wads):
        """Create a guesser from wads"""
        hashes = set(wf.path_hash for wad in wads for wf in wad.files)
        self = cls(hashes)
        self.wads = wads
        return self

    @staticmethod
    def unknown_from_export(path):
        """Load unknown hashes from 'export/*.unknown.txt' files"""

        unknown = set()
        for p in glob.glob(f"{path}/*.unknown.txt"):
            with open(p) as f:
                unknown |= {int(h, 16) for h in f}
        return unknown

    def save(self):
        self.hashfile.save()

    def _add_known(self, h, p):
        print("%016x %s" % (h, p))
        self.known[h] = p
        self.unknown.remove(h)

    def check(self, p):
        """Check a single hash, print and add to known on match"""

        h = xxh64_intdigest(p)
        if h in self.unknown:
            self._add_known(h, p)

    def is_known(self, p):
        """Check a path, return True if it is known"""

        h = xxh64_intdigest(p)
        if h in self.unknown:
            self._add_known(h, p)
            return True
        return h in self.known

    def check_iter(self, paths):
        """Check paths from an iterable"""

        # failsafe for common dumb error
        if isinstance(paths, str):
            raise TypeError("expected iterable of strings, got a string")
        unknown = self.unknown
        for p in paths:
            h = xxh64_intdigest(p)
            if h in unknown:
                self._add_known(h, p)

    def check_text_list(self, text):
        """Check paths from a text list"""
        self.check_iter(s for s in text.split() if s)

    def check_xdbg_hashes(self, path):
        with open(path) as f:
            self.check_iter(l.split('"')[1] for l in f if l.startswith('hash: '))

    def check_basenames(self, names):
        """Check a list of basenames for each known subdirectory"""

        dirs = self.directory_list()
        for name in progress_iterator(sorted(names)):
            self.check_iter(f"{dir}/{name}" for dir in dirs)

    def directory_list(self, cached=True):
        """Return a set of all directories and subdirectories"""

        if not cached or self.__directory_list is None:
            # do multiple passes to return intermediate directories
            dirs = set()
            bases = self.known.values()
            while len(bases):
                bases = {os.path.dirname(p) for p in bases} - dirs
                dirs |= bases
            self.__directory_list = list(dirs)
        return self.__directory_list

    def substitute_basenames(self):
        """Check all basenames in each subdirectory"""

        names = {os.path.basename(p) for p in self.known.values()}
        dirs = self.directory_list()
        logger.debug(f"substitute basenames: {len(names)} basenames, {len(dirs)} directories")
        for name in progress_iterator(sorted(names)):
            self.check_iter(f"{dir}/{name}" for dir in dirs)

    def _substitute_basename_words(self, paths, words, nold=1, nnew=1):
        """Replaces nold side by side words with nnew words in all basenames of all given paths
        (default=1 for simple 1 word substitution). nold > 0; nnew > 0 is required."""

        if not (nold and nnew):
            raise ValueError("Either nold or nnew is 0.")
        words = list(words) # Ensure words is a list
        format_part = "{sep}%%s" * (nnew-1)
        format_part = f"%s%%s{format_part}%s"
        re_extract = re.compile(rf"([^/_.-]+)(?=((?:[-_][^/_.-]+){{{nold-1}}})[^/]*\.[^/]+$)")
        temp_formats = set()
        for path in paths:
            for m in re_extract.finditer(path):
                match = m.group(1) + m.group(2)
                temp_formats.add(format_part % (path[:m.start()], path[m.span()[0]+len(match):]))

        formats = {fmt.replace("{sep}", sep) for fmt in temp_formats for sep in "-_"}

        product = itertools.product
        logger.debug(f"substitute basename words ({nold} by {nnew}): {len(formats)} formats, {len(words)} words")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt % p for p in product(words, repeat=nnew))

    def _add_basename_word(self, paths, words):
        """Add a word to all known basenames"""

        words = list(words) # Ensure words is a list
        re_extract = re.compile(r'([^/_.-]+)(?=[^/]*\.[^/]+$)')
        formats = set()
        for path in paths:
            for m in re_extract.finditer(path):
                formats.update('%s%%s%s%s' % (path[:m.start()], sep, path[m.start():]) for sep in "-_")
                formats.update('%s%s%%s%s' % (path[:m.end()], sep, path[m.end():]) for sep in "-_")

        logger.debug(f"add basename word: {len(formats)} formats, {len(words)} words")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt % w for w in words)

    def _substitute_numbers(self, paths, nmax=10000, digits=None):
        """Guess hashes by changing numbers in basenames"""

        if digits is True:
            # guess digits count from nmax
            digits = len(str(nmax - 1))
        if digits is None:
            fmt = '%d'
            re_extract = re.compile(r'([0-9]+)(?=[^/]*\.[^/]+$)')
        else:
            fmt = f"%0{digits}d"
            re_extract = re.compile(r'([0-9]{%d})(?=[^/]*\.[^/]+$)' % digits)

        formats = set()
        for path in paths:
            for m in re_extract.finditer(path):
                formats.add('%s%s%s' % (path[:m.start()], fmt, path[m.end():]))

        nrange = range(nmax)
        logger.debug(f"substitute numbers: {len(formats)} formats, nmax = {nmax}")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt % n for n in nrange)

    def substitute_extensions(self):
        """Guess hashes by substituting file extensions"""

        prefixes = set()
        extensions = set()
        for path in self.known.values():
            prefix, ext = os.path.splitext(path)
            prefixes.add(prefix)
            extensions.add(ext)

        logger.debug(f"substitute extensions: {len(prefixes)} prefixes, {len(extensions)} extensions")
        for prefix in progress_iterator(sorted(prefixes)):
            self.check_iter(prefix + ext for ext in extensions)

    def wad_text_files(self, wad):
        """Iterate over wad files, generate text file data"""

        with open(wad.path, 'rb') as f:
            for wadfile in wad.files:
                # skip non-text files as soon as possible
                if wadfile.ext in ('png', 'jpg', 'ttf', 'webm', 'ogg', 'dds', 'tga'):
                    continue
                try:
                    data = wadfile.read_data(f).decode('utf-8-sig')
                except UnicodeDecodeError:
                    continue
                if data:
                    yield wadfile, data


class LcuHashGuesser(HashGuesser):
    def __init__(self, hashes):
        super().__init__(hashfile_lcu, hashes)

    @classmethod
    def from_wads(cls, wads):
        """Create a guesser from wads

        WADs whose extension is not .wad are ignored.
        """
        return super().from_wads([wad for wad in wads if wad.path.endswith('.wad')])

    def build_wordlist(self):
        re_filter_path = re.compile(r'(?:^plugins/rcp-be-lol-game-data/global/default/data/characters/|/[0-9a-f]{32}\.)')
        paths = (p for p in self.known.values() if not re_filter_path.search(p))
        return build_wordlist(paths)

    def substitute_region_lang(self):
        """Guess hashes from region/lang variants"""

        regions = REGIONS + ['global']
        langs = [l.value for l in Language] + ['default']

        regex = re.compile(r'^plugins/([^/]+)/[^/]+/[^/]+/')
        region_lang_list = [(r, l) for r in regions for l in langs]
        known = list(self.known.values())
        logger.debug("substitute region and lang")
        for region_lang in progress_iterator(region_lang_list, lambda rl: f"{rl[0]}/{rl[1]}"):
            replacement = r'plugins/\1/%s/%s/' % region_lang
            self.check_iter(regex.sub(replacement, p) for p in known)

    def substitute_basename_words(self, plugin=None, fileext=None, words=None, nold=1, nnew=1):
        """Replaces nold side by side words in basenames with nnew words (default=1 for simple 1 word substitution).
        Additionally, a plugin name, file extension and wordlist can be specified to filter on."""

        if words is None:
            words = self.build_wordlist()
        paths = self.known.values()
        if plugin:
            paths = [path for path in paths if path.startswith(f"plugins/{plugin}/")]
        if fileext:
            paths = [path for path in paths if path.endswith(fileext)]

        super()._substitute_basename_words(paths, words, nold=nold, nnew=nnew)

    def add_basename_word(self):
        super()._add_basename_word(self.known.values(), self.build_wordlist())

    def substitute_numbers(self, nmax=10000, digits=None):
        re_filter = re.compile(r"""(?:
            ^(?:plugins/rcp-be-lol-game-data/[^/]+/[^/]+/v1/champion-
              | plugins/rcp-be-lol-game-data/global/default/(?:data|assets)/characters/
              | plugins/rcp-be-lol-game-data/global/default/data/items/icons2d/\d+_
              | plugins/rcp-be-lol-game-data/[^/]+/[^/]+/v1/champions/-1.json
              )
            | /[0-9a-f]{32}\.
            )""", re.VERBOSE)
        paths = (p for p in self.known.values() if not re_filter.search(p))
        super()._substitute_numbers(paths, nmax, digits)

    def substitute_plugin(self):
        """Guess hashes by changing plugin name"""

        all_paths = [p for p in self.known.values() if p.startswith('plugins/')]
        plugins = {p.split('/', 2)[1] for p in all_paths}
        formats = {re.sub(r'^plugins/([^/]+)/', r'plugins/%s/', p) for p in all_paths}

        logger.debug(f"substitute plugin: {len(formats)} formats, {len(plugins)} plugins")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt % p for p in plugins)

    def grep_wad(self, wad):
        """Find hashes from a wad file"""

        logger.debug(f"find LCU hashes in WAD {wad.path}")
        # candidate relative subpaths (not lowercased yet)
        relpaths = set()

        for wadfile, data in self.wad_text_files(wad):
            jdata = None
            if wadfile.ext == 'json':
                # parse specific information from known json files
                try:
                    jdata = json.loads(data)
                except (json.JSONDecodeError, UnicodeDecodeError):
                    pass

            if jdata is not None:
                if wadfile.path == 'plugins/rcp-fe-lol-loot/global/default/trans.json':
                    # names of hextech items, keys are also image names
                    self.check_iter(f"plugins/rcp-be-lol-game-data/global/default/v1/hextech-images/{k}.png" for k in jdata)
                    continue  # no more data to parse
                elif 'pluginDependencies' in jdata and 'name' in jdata:
                    # retrieve plugin name from description.json
                    # guess some common paths
                    name = jdata['name']
                    subpaths = ['index.html', 'init.js', 'init.js.map', 'bundle.js', 'trans.json', 'css/main.css', 'license.json']
                    self.check_iter(f"plugins/{name}/global/default/{subpath}" for subpath in subpaths)
                elif 'musicVolume' in jdata and 'files' in jdata:
                    # splash config
                    # try to guess subdirectory name (names should only contain one element)
                    names = {s.lower() for path in jdata['files'].values() for s in re.findall(r'-splash-([^.]+)', path)}
                    self.check_iter(f"plugins/rcp-fe-lol-splash/global/default/splash-assets/{name}/config.json" for name in names)
                    self.check_iter(f"plugins/rcp-fe-lol-splash/global/default/splash-assets/{name}/{path.lower()}" for name in names for path in jdata['files'].values())
                    continue  # no more data to parse
                elif wadfile.path == 'plugins/rcp-be-lol-game-data/global/default/v1/champion-summary.json':
                    champion_ids = [v['id'] for v in jdata]
                    self.check_iter(f'plugins/rcp-be-lol-game-data/global/default/v1/champions/{cid}.json' for cid in champion_ids)
                    self.check_iter(f'plugins/rcp-be-lol-game-data/global/default/v1/champion-splashes/{cid}/metadata.json' for cid in champion_ids)
                elif 'recommendedItemDefaults' in jdata:
                    # plugins/rcp-be-lol-game-data/global/default/v1/champions/{cid}.json
                    self.check_iter(f'plugins/rcp-be-lol-game-data/global/default{p.lower()}' for p in jdata['recommendedItemDefaults'])

            # search for known paths formats
            # /fe/{plugin}/{subpath} -> plugins/rcp-fe-{plugin}/global/default/{subpath}
            self.check_iter(f"plugins/rcp-fe-{m.group(1)}/global/default/{m.group(2)}".lower()
                            for m in re.finditer(r"\bfe/([^/]+)/([a-zA-Z0-9/_.@-]+)", data))
            # /DATA/{subpath} -> plugins/rcp-be-lol-game-data/global/default/data/{subpath}
            self.check_iter(f"plugins/rcp-be-lol-game-data/global/default/data/{m.group(1)}".lower()
                            for m in re.finditer(r"/DATA/([a-zA-Z0-9/_.@-]+)", data))
            # /lol-game-data/assets/{subpath} -> plugins/rcp-be-lol-game-data/global/default/{subpath}
            self.check_iter(f"plugins/rcp-be-lol-game-data/global/default/{m.group(1)}".lower()
                            for m in re.finditer(r'\blol-game-data/assets/([a-zA-Z0-9/_.@-]+)', data))

            # relative path starting with ./ or ../ (e.g. require() use)
            relpaths |= {m.group(1) for m in re.finditer(r'[^a-zA-Z0-9/_.\\-]((?:\.|\.\.)/[a-zA-Z0-9/_.-]+)', data)}
            # basename or subpath (check for an extension)
            relpaths |= {m.group(1) for m in re.finditer(r'''["']([a-zA-Z0-9][a-zA-Z0-9/_.@-]*\.(?:js|json|webm|html|[a-z]{3}))\b''', data)}
            # template ID to template path
            relpaths |= {f"{m.group(1)}/template.html" for m in re.finditer(r'<template id="[^"]*-template-([^"]+)"', data)}
            # JS maps
            relpaths |= {m.group(1) for m in re.finditer(r'sourceMappingURL=(.*?\.js)\.map', data)}

        self.check_basenames(p.lower() for p in relpaths)

    def guess_from_game_hashes(self):
        """Guess LCU hashes from game hashes"""

        base = 'plugins/rcp-be-lol-game-data/global/default'
        for path in hashfile_game.load().values():
            prefix, ext = os.path.splitext(path)
            if ext == '.dds':
                self.check(f"{base}/{prefix}.png")
                self.check(f"{base}/{prefix}.jpg")
            elif ext == '.json':
                self.check(f"{base}/{path}")

    def guess_patterns(self):
        """Guess from known path patterns"""

        # note: most values are already retrieved from json files in lol-game-data WAD

        # runes (perks)
        #XXX valid values, including 'p{i}_s{j}_k{k}.jpg' basenames could be retrieved from perkstyles.json
        perk_primary = range(8000, 8600, 100)
        for i in perk_primary:
            perk_secondary = range(i, i + 100)
            self.check_iter(f'plugins/rcp-fe-lol-perks/global/default/images/inventory-card/{i}/p{i}_s{j}_k{k}.jpg'
                            for j in [0] + list(perk_primary)
                            for k in [0] + list(perk_secondary)
                            )
            paths = ['environment.jpg', 'construct.png']
            paths += [f'keystones/{j}.png' for j in perk_secondary]
            paths += [f'second/{j}.png' for j in perk_primary]
            self.check_iter(f"plugins/rcp-fe-lol-perks/global/default/images/construct/{i}/{p}" for p in paths)

        # sanitizer
        langs = [l.value for l in Language]
        paths = []
        for i in range(5):
            for action in ('filter', 'unfilter', 'whitelist'):
                paths += [f'{i}.{action}.csv']
                paths += [f'{i}.{action}.language.{x.split("_")[0]}.csv' for x in langs]
                paths += [f'{i}.{action}.country.{x.split("_")[1]}.csv' for x in langs]
                paths += [f'{i}.{action}.region.{x}.csv' for x in REGIONS]
                paths += [f'{i}.{action}.locale.{x}.csv' for x in langs]
        for p in 'allowedchars breakingchars projectedchars projectedchars1337 punctuationchars variantaliases'.split():
            paths += [f'{p}.locale.{x}.txt' for x in langs]
            paths += [f'{p}.language.{x.split("_")[0]}.txt' for x in langs]
        self.check_iter(f'plugins/rcp-be-sanitizer/global/default/{p}' for p in paths)

        # plugins/rcp-fe-lol-loot/global/default/assets/loot_item_icons/{name}.png -> {name}_splash.png
        for p in self.known.values():
            if p.startswith('plugins/rcp-fe-lol-loot/global/default/assets/loot_item_icons/'):
                self.check(p.replace('.png', '_splash.png'))

        # patterns already checked when substituting numbers (and sometimes retrieved from WAD files)
        #  plugins/rcp-be-lol-game-data/global/default/content/src/leagueclient/wardskinimages/wardhero_{i}.png
        #  plugins/rcp-be-lol-game-data/global/default/content/src/leagueclient/wardskinimages/wardheroshadow_{i}.png
        #  plugins/rcp-be-lol-game-data/global/default/v1/profile-icons/{i}.jpg
        #  plugins/rcp-be-lol-game-data/global/default/v1/summoner-backdrops/{i}.jpg
        #  plugins/rcp-be-lol-game-data/global/default/v1/summoner-backdrops/{i}.webm
        #  plugins/rcp-fe-lol-loot/global/default/assets/loot_item_icons/chest_{i}.png
        #  plugins/rcp-fe-lol-loot/global/default/assets/loot_item_icons/chest_{i}_open.png
        #  plugins/rcp-fe-lol-loot/global/default/assets/loot_item_icons/material_{i}.png
        #  plugins/rcp-be-lol-game-data/global/default/v1/hextech-images/champion_skin_{i}.png
        #  plugins/rcp-be-lol-game-data/global/default/v1/hextech-images/champion_skin_rental_{i}.png
        #  plugins/rcp-fe-lol-skins-viewer/global/default/video/collection/{i}.webm


class GameHashGuesser(HashGuesser):
    def __init__(self, hashes):
        super().__init__(hashfile_game, hashes)

    @classmethod
    def from_wads(cls, wads):
        """Create a guesser from wads

        WADs whose extension is not .wad.client are ignored.
        """
        return super().from_wads([wad for wad in wads if wad.path.endswith('.wad.client')])

    def build_wordlist(self):
        return build_wordlist(self.known.values())

    def get_characters(self):
        re_char = re.compile(r'^(?:assets|data)/characters/([^/]+)/')
        chars = set()
        for p in self.known.values():
            m = re_char.match(p)
            if m:
                chars.add(m.group(1))
        return sorted(chars)

    shader_extensions = [".ps_2_0", ".ps_3_0", ".vs_2_0", ".vs_3_0", ".ps", ".vs"]
    shader_variants = [".dx11", ".dx9", ".dx9sm3", ".glsl", ".metal"]

    def substitute_numbers(self, nmax=100, digits=None):
        paths = self.known.values()
        super()._substitute_numbers(paths, nmax, digits)

    def check_basename_prefixes(self, prefixes=None):
        """Checks a provided list of prefixes for all basenames.
        If no list is provided, a default one will be used"""

        values = set()
        if prefixes is None:
            prefixes = ['2x_', '2x_sd_', '4x_', '4x_sd_', 'sd_']
        for p in self.known.values():
            path, basename = os.path.split(p)
            if path:
                path += '/'
            values.update(f"{path}{prefix}{basename}" for prefix in prefixes)

        logger.debug(f"check basename prefixes: {len(prefixes)} prefixes with a total {len(values)} paths")
        self.check_iter(value for value in list(values))

    def substitute_basename_words(self):
        super()._substitute_basename_words(self.known.values(), self.build_wordlist())

    def add_basename_word(self):
        paths = [path for path in self.known.values() if not any(part in path for part in ['assets/characters/', 'vo/', 'sfx/', 'skins_skin'])]
        super()._add_basename_word(paths, self.build_wordlist())

    def substitute_character(self):
        """Guess hashes by changing champion names in assets/characters/"""

        characters = set()
        formats = set()
        re_char = re.compile(r'^(?:assets|data)/characters/([^/]+)/')
        for p in self.known.values():
            m = re_char.match(p)
            if not m:
                continue
            char = m.group(1)
            characters.add(char)
            formats.add(p.replace(char, '{}'))

        logger.debug(f"substitute characters: {len(formats)} formats, {len(characters)} characters")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt.replace('{}', s) for s in characters)

    def substitute_skin_numbers(self):
        """Replace skinNN, multiple combinations"""

        characters = {}  # {char: ({skin}, {(format, N})}
        regex = re.compile(r'/characters/([^/]+)/skins/(base|skin\d+)/')
        for p in self.known.values():
            m = regex.search(p)
            if not m:
                continue
            char, skin = m.groups()
            if m.group(1) == 'sightward':
                continue
            c = characters.setdefault(char, (set(), set()))
            c[0].add(skin)
            c[1].add(re.subn(r'(?:base|skin\d+)', '%s', p))

        # generate all combinations
        logger.debug(f"substitute skin numbers: {len(characters)} characters")
        for char, (skins, formats) in progress_iterator(characters.items(), lambda v: v[0]):
            for fmt, nocc in formats:
                self.check_iter(fmt % p for p in itertools.combinations(skins, nocc))

    def substitute_suffixes(self):
        """Replace `.suffix.ext` using all known suffixes"""

        suffixes = {""}
        formats = set()
        re_suffix = re.compile(r'^(.*?)(\.[^.]+)?(\.[^.]+)$')
        for p in self.known.values():
            m = re_suffix.search(p)
            if not m:
                continue
            prefix, suffix, ext = m.groups()
            if suffix:
                suffixes.add(suffix)
            formats.add(f"{prefix}%s{ext}")

        # generate all combinations
        logger.debug(f"substitute suffixes: {len(formats)} formats, {len(suffixes)} suffixes")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt % s for s in suffixes)

    def substitute_lang(self):
        """Guess hashes from lang variants"""

        langs = [l.value for l in Language]
        langs_re = re.compile(r'(%s)' % '|'.join(langs))
        formats = {langs_re.sub('{}', p) for p in self.known.values() if langs_re.search(p)}

        logger.debug(f"substitute lang: {len(formats)} formats, {len(langs)} langs")
        for fmt in progress_iterator(sorted(formats)):
            self.check_iter(fmt.replace('{}', l) for l in langs)

    def guess_skin_groups_bin_using_chromas(self):
        """Guess 'skin*.bin' with long filenames using chroma groups"""

        # retrieve chromas from cdragon
        import requests
        data = requests.get('http://raw.communitydragon.org/pbe/plugins/rcp-be-lol-game-data/global/default/v1/skins.json').json()

        char_to_skin_groups = {}
        char_name_re = re.compile(r'^/lol-game-data/assets/assets/characters/([^/]+)/skins/')
        for skin_id, skin_data in data.items():
            char_name = char_name_re.match(skin_data['loadScreenPath'].lower()).group(1)

            group_skin_ids = [skin_id]
            if skin_data.get('chromas'):
                group_skin_ids.extend(d['id'] for d in skin_data['chromas'])
            char_to_skin_groups.setdefault(char_name, []).append([int(i) % 1000 for i in group_skin_ids])

        logger.debug("find skin groups .bin files using chroma groups")
        for char, groups in progress_iterator(char_to_skin_groups.items(), lambda v: v[0]):
            str_groups = [[f"_skins_skin{i}" for i in group] for group in groups] + [["_skins_root"]]
            for n in range(len(str_groups)):
                for p in itertools.combinations(str_groups, n+1):
                    # note: skins are in lexicographic order: skin11 is before skin2; root before all
                    s = ''.join(sorted(s for g in p for s in g))
                    self.check(f"data/{char}{s}.bin")

    def guess_skin_groups_bin(self):
        """Guess 'skin*.bin' with long filenames"""

        char_to_skins = {}
        regex = re.compile(r'^assets/characters/([^/]+)/skins/skin(\d+)/')
        for p in self.known.values():
            m = regex.match(p)
            if not m:
                continue
            char = m.group(1)
            nskin = int(m.group(2))
            if char == 'sightward':
                continue
            # always add skin 0 (base skin)
            char_to_skins.setdefault(char, {0}).add(nskin)

        # generate all combinations
        logger.debug("find skin groups .bin files")
        for char, skins in progress_iterator(char_to_skins.items(), lambda v: v[0]):
            # note: skins are in lexicographic order: skin11 is before skin2
            str_skins = sorted(f"_skins_skin{i}" for i in skins)
            for n in range(len(str_skins)):
                for p in itertools.combinations(str_skins, n+1):
                    s = ''.join(p)
                    self.check(f"data/{char}{s}.bin")

    def guess_from_lcu_hashes(self):
        """Guess game hashes from LCU hashes"""

        re_data = re.compile(r"^plugins/rcp-be-lol-game-data/global/default/((?:assets|data)/.*)\.(png|jpg|json)$")
        for path in hashfile_lcu.load().values():
            m = re_data.match(path)
            if not m:
                continue
            prefix, ext = m.groups()
            if ext in ('png', 'jpg'):
                self.check(f"{prefix}.dds")
            else:
                self.check(f"{prefix}.{ext}")

    def guess_characters_files(self, chars=None):
        """Guess hashes using common patterns for characters files"""

        if chars is None:
            chars = self.get_characters()

        formats = [
            "data/characters/{c}/skins/root.bin",
            "data/characters/{c}/skins/base/{c}.skl",
            "data/characters/{c}/skins/base/{c}.skn",
            "data/characters/{c}/skins/base/{c}_tx_cm.dds",
            "data/characters/{c}/tiers/root.bin",
            "data/characters/{c}/{c}.bin",
            "data/characters/{c}/{c}.ddf",
            "data/characters/{c}/hud/{c}_circle.dds",
            "data/characters/{c}/hud/{c}_square.dds",
            "assets/characters/{c}/hud/{c}_circle.dds",
            "assets/characters/{c}/hud/{c}_square.dds",
            "characters/{c}"
        ]

        logger.debug(f"guess characters files: {len(chars)} characters")
        for c in progress_iterator(sorted(chars)):
            self.check_iter(s.format(c=c) for s in formats)
            nskins = 500 if c == 'sightward' else 200
            self.check_iter(f"data/characters/{c}/skins/skin{i}.bin" for i in range(nskins))
            self.check_iter(f"data/characters/{c}/animations/skin{i}.bin" for i in range(nskins))
            if c.startswith('pet'):
                self.check_iter(f"data/characters/{c}/tiers/tier{i}.bin" for i in range(10))

    def guess_shader_variants(self):
        """Guess different extension variants for shader files, e.g. ".glsl_100" """

        shader_pattern = re.compile(r'.*\.[pv]s(_[23]_0|(?=$|\.))')
        shader_paths = {match.group(0) for path in self.known.values() if (match := shader_pattern.match(path)) is not None}
        for path in sorted(shader_paths):
            self.check_iter(f"{path}{variant}" for variant in self.shader_variants)
            self.check_iter(f"{path}{variant}_{n}" for variant in self.shader_variants for n in range(0, 20000, 100))

    def grep_wad(self, wad):
        """Find hashes from a wad file"""

        logger.debug(f"find game hashes in WAD {wad.path}")

        with open(wad.path, 'rb') as f:
            for wadfile in wad.files:
                if wadfile.type == 2:
                    continue # softlink; contains no actual content
                if wadfile.ext in ('dds', 'jpg', 'png', 'tga', 'ttf', 'otf', 'ogg', 'webm', 'anm',
                                   'skl', 'skn', 'scb', 'sco', 'troybin', 'bnk', 'wpk', 'tex'):
                    continue # don't grep filetypes known to not contain full paths

                data = wad.read_file_data(f, wadfile)
                if data is None:
                    continue
                if wadfile.ext in ('bin', 'inibin'):
                    # bin files: find strings based on prefix, then parse the length
                    for m in re.finditer(br'(?:ASSETS|DATA|Characters|Shaders|Maps/MapGeometry|Gameplay|ClientStates|Patching)/', data):
                        i = m.start()
                        n = data[i-2] + (data[i-1] << 8)
                        try:
                            path = data[i:i+n].lower().decode('ascii')
                        except UnicodeDecodeError:
                            continue
                        if path.startswith('characters'):
                            self.check(f"assets/{path}")
                            self.check(f"data/{path}")
                        elif path.endswith('.lua'):
                            self.check(path[:-4] + '.luabin')
                            self.check(path[:-4] + '.luabin64')
                        elif path.startswith('shaders'):
                            self.check_iter(f"assets/shaders/generated/{path}{ext}" for ext in self.shader_extensions)
                            self.check_iter(f"assets/shaders/generated/{path}{ext}{variant}" for ext in self.shader_extensions for variant in self.shader_variants)
                        elif path.startswith('maps'):
                            self.check(f"data/{path}.mapgeo")
                            self.check(f"data/{path}.materials.bin")
                        elif path.startswith('clientstates') or path.startswith('patching'):
                            self.check(path.rsplit('/', 1)[0])
                            self.check(path.rsplit('/', 2)[0])
                        else:
                            self.check(path)
                            if path.endswith(".png"):
                                self.check(path[:-4] + ".dds")

                elif wadfile.ext == 'preload':
                    # preload files
                    for m in re.finditer(br'Name="([^"]+)"', data):
                        path = m.group(1).lower().decode('ascii')
                        if path.endswith('.lua'):
                            self.check(path[:-4] + '.luabin')
                            self.check(path[:-4] + '.luabin64')
                        elif path.endswith('.troy'):
                            self.check('data/shared/particles/'+ path[:-5] + '.troybin')
                        elif wadfile.path:
                            fmt = os.path.dirname(wadfile.path) + '/%s.preload'
                            self.check(fmt % path)

                elif wadfile.ext in ('hls', 'ps_2_0', 'ps_3_0', 'vs_2_0', 'vs_3_0'):
                    # shader: search for includes
                    if wadfile.path:
                        dirname = os.path.dirname(wadfile.path)
                        for m in re.finditer(br'#include "([^"]+)"', data):
                            subpath = m.group(1).lower().decode('ascii')
                            self.check(os.path.normpath(f"{dirname}/{subpath}"))

                else:
                    # fallback: search for path-looking strings in all remaining files
                    self.grep_file(data=data)

    def grep_file(self, path=None, data=None):
        if path:
            with open(path, 'rb') as f:
                data = f.read()
        elif data is None:
            raise TypeError("either path or data must be provided")

        # find path-like strings, then try to parse the length
        paths = set()
        for m in re.finditer(br'(?:ASSETS|Common|DATA|DATA_SOON|DATA_Soon|Gameplay|Global|LEVELS|Loadouts|UX|UIAutoAtlas)/[0-9a-zA-Z_. /-]+', data):
            path = m.group(0).lower().decode('ascii')
            paths.add(path.replace("data_soon/", "data/"))
            pos = m.start()
            if pos >= 2:
                n = struct.unpack('<H', data[pos-2:pos])[0]
                if n == 0 and pos >= 4:
                    n = struct.unpack('<L', data[pos-4:pos])[0]
                if n < len(path):
                    paths.add(path[:n].replace("data_soon/", "data/"))

        for p in paths:
            if p.endswith('.lua'):
                self.check(p[:-4] + '.luabin')
                self.check(p[:-4] + '.luabin64')
            else:
                self.check(p)



import logging
logger = logging.getLogger(__name__)

from cdtb.storage import (
    Storage,
    Patch,
    PatchElement,
)
from cdtb.wad import (
    Wad,
)
from cdtb.export import (
    Exporter,
    CdragonRawPatchExporter,
)

# import storages to register them
import cdtb.rads
import cdtb.patcher




import os
from enum import IntEnum
import struct
from .hashes import HashFile, default_hash_dir, hashfile_game
from xxhash import xxh64_intdigest


def _repr_indent(v):
    return repr(v).replace('\n', '\n  ')

def _repr_indent_list(values):
    if not values:
        return '[]'
    return "[\n%s]" % ''.join(f"  {_repr_indent(v)}\n" for v in values)


hashfile_binentries = HashFile(default_hash_dir / "hashes.binentries.txt", hash_size=8)
hashfile_binhashes = HashFile(default_hash_dir / "hashes.binhashes.txt", hash_size=8)
hashfile_binfields = HashFile(default_hash_dir / "hashes.binfields.txt", hash_size=8)
hashfile_bintypes = HashFile(default_hash_dir / "hashes.bintypes.txt", hash_size=8)
hashfile_binpaths = hashfile_game

def compute_binhash(s):
    """Compute a hash used in BIN files

    FNV-1a hash, on lowercased input
    """
    h = 0x811c9dc5
    for b in s.encode('ascii').lower():
        h = ((h ^ b) * 0x01000193) % 0x100000000
    return h

class BinHashBase:
    """Base class for hashed value"""

    hashfile = None  # to be defined in subclasses

    def __init__(self, h):
        self.h = h
        self.s = self.hashfile.load().get(h)

    def __eq__(self, other):
        if isinstance(other, BinHashBase):
            return self.h == other.h
        elif isinstance(other, str):
            if self.s is None:
                return self.h == self.compute_hash(other)
            else:
                return self.s == other
        else:
            return self.h == other

    def __str__(self):
        if self.s is not None:
            return self.s
        return f"{{{self.hex()}}}"

    __repr__ = __str__
    to_serializable = __str__

    def __hash__(self):
        return self.h

    @classmethod
    def compute_hash(cls, s):
        return compute_binhash(s)

    def hex(self):
        return f"{self.h:08x}"

class BinHashValue(BinHashBase):
    """Hashed name in bin files (hash type)"""

    hashfile = hashfile_binhashes

    def __repr__(self):
        if self.s is not None:
            return repr(self.s)
        return f"{{{self.hex()}}}"

class BinEntryPath(BinHashBase):
    """Path of a bin entry (top level element)"""

    hashfile = hashfile_binentries

    def __repr__(self):
        if self.s is not None:
            return repr(self.s)
        return f"{{{self.hex()}}}"

class BinFieldName(BinHashBase):
    """Name of a struct field"""

    hashfile = hashfile_binfields

class BinTypeName(BinHashBase):
    """Name of a type"""

    hashfile = hashfile_bintypes

class BinPathValue(BinHashBase):
    """Hashed WAD path in bin files"""

    hashfile = hashfile_binpaths

    def hex(self):
        return f"{self.h:016x}"

    @classmethod
    def compute_hash(cls, s):
        return xxh64_intdigest(s.lower())

    def __repr__(self):
        if self.s is not None:
            return repr(self.s)
        return f"{{{self.hex()}}}"


def key_to_hash(key):
    if isinstance(key, BinHashBase):
        return key.h
    elif isinstance(key, str):
        return compute_binhash(key)
    else:
        return key

class BinObjectWithFields:
    """Base class for bin object with fields"""

    def __init__(self, fields):
        self.fields = fields

    def __getitem__(self, key):
        h = key_to_hash(key)
        for v in self.fields:
            if v.name.h == h:
                return v
        raise KeyError(key)

    def __setitem__(self, key, value):
        h = key_to_hash(key)
        for n, v in enumerate(self.fields):
            if v.name.h == h:
                self.fields[n] = value
                break
        else:
            self.fields.append(value)

    def __contains__(self, key):
        h = key_to_hash(key)
        for v in self.fields:
            if v.name.h == h:
                return True
        return False

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def getv(self, key, default=None):
        try:
            return self[key].value
        except KeyError:
            return default

    def get_path(self, *args, default=None):
        """Get value by accessing successive field names"""
        v = self
        try:
            for key in args:
                v = v[key].value
            return v
        except KeyError:
            return default

    def to_serializable(self):
        return dict(f.to_serializable() for f in self.fields)

class BinObjectWithFieldsAndType(BinObjectWithFields):
    """Base class for bin objects with fields and type"""

    def __init__(self, htype, fields):
        super().__init__(fields)
        self.type = BinTypeName(htype)

    def to_serializable(self):
        serialized = super().to_serializable()
        serialized["__type"] = self.type.to_serializable()
        return serialized

class BinType(IntEnum):
    # See parse_bintype() for remapping depending on version
    EMPTY = 0
    BOOL = 1
    S8 = 2
    U8 = 3
    S16 = 4
    U16 = 5
    S32 = 6
    U32 = 7
    S64 = 8
    U64 = 9
    FLOAT = 10
    VEC2_FLOAT = 11
    VEC3_FLOAT = 12
    VEC4_FLOAT = 13
    MATRIX4X4 = 14
    RGBA = 15
    STRING = 16
    HASH = 17
    PATH = 18  # introduced in 10.23
    # Complex types (0x80 flag introduced in 9.23)
    CONTAINER = 0x80
    CONTAINER2 = 0x81  # introduced in 10.8
    STRUCT = 0x82
    EMBEDDED = 0x83
    LINK = 0x84
    OPTION = 0x85
    MAP = 0x86
    FLAG = 0x87


class BinStruct(BinObjectWithFieldsAndType):
    """Structured binary value"""

    def __repr__(self):
        sfields = _repr_indent_list(self.fields)
        return f"<STRUCT {self.type!r} {sfields}>"

class BinEmbedded(BinObjectWithFieldsAndType):
    """Embedded binary value"""

    def __repr__(self):
        sfields = _repr_indent_list(self.fields)
        return f"<EMBEDDED {self.type!r} {sfields}>"

class BinNested(BinObjectWithFields):
    """Nested binary value"""

    def __repr__(self):
        sfields = _repr_indent_list(self.fields)
        return f"<{sfields}>"

class BinField:
    """Base class for binary fields

    A field is a value (possibly nested) associated to a hash.
    """
    def __init__(self, hname):
        self.name = BinFieldName(hname)

class BinBasicField(BinField):
    """Binary field for fixed-width, non-nested values"""

    def __init__(self, hname, btype, value):
        super().__init__(hname)
        self.type = btype
        self.value = value

    def __repr__(self):
        return f"<{self.name!r} {self.type.name} {self.value!r}>"

    def to_serializable(self):
        return (self.name.to_serializable(), _to_serializable(self.value))

class BinContainerField(BinField):
    def __init__(self, hname, btype, values):
        super().__init__(hname)
        self.type = btype
        self.value = values

    def __repr__(self):
        svalues = _repr_indent_list(self.value)
        return f"<{self.name!r} CONTAINER({self.type.name}) {svalues}>"

    def to_serializable(self):
        return (self.name.to_serializable(), [_to_serializable(v) for v in self.value])

class BinStructField(BinField):
    def __init__(self, hname, value):
        super().__init__(hname)
        self.value = value

    def __repr__(self):
        sfields = _repr_indent_list(self.value.fields)
        return f"<{self.name!r} STRUCT {self.value.type!r} {sfields}>"

    def to_serializable(self):
        return (self.name.to_serializable(), self.value.to_serializable())

class BinEmbeddedField(BinField):
    def __init__(self, hname, value):
        super().__init__(hname)
        self.value = value

    def __repr__(self):
        sfields = _repr_indent_list(self.value.fields)
        return f"<{self.name!r} EMBEDDED {self.value.type!r} {sfields}>"

    def to_serializable(self):
        return (self.name.to_serializable(), self.value.to_serializable())

class BinOptionField(BinField):
    def __init__(self, hname, vtype, value):
        super().__init__(hname)
        self.vtype = vtype
        self.value = value

    def __repr__(self):
        svalue = '-' if self.value is None else f'(\n  {_repr_indent(self.value)}\n)'
        return f"<{self.name!r} OPTION({self.vtype.name}) {svalue}>"

    def to_serializable(self):
        return (self.name.to_serializable(), None if self.value is None else _to_serializable(self.value))

class BinMapField(BinField):
    def __init__(self, hname, ktype, vtype, values):
        super().__init__(hname)
        self.ktype = ktype
        self.vtype = vtype
        self.value = values

    def __repr__(self):
        svalues = ''.join(f"  {k} => {_repr_indent(v)}\n" for k, v in self.value.items())
        return f"<{self.name!r} MAP({self.ktype.name},{self.vtype.name}) {{\n{svalues}}}>"

    def to_serializable(self):
        return (self.name.to_serializable(), {_to_serializable(k): _to_serializable(v) for k,v in self.value.items()})

class BinNestedField(BinField):
    def __init__(self, hname, value):
        super().__init__(hname)
        self.value = value

    def __repr__(self):
        sfields = _repr_indent_list(self.value.fields)
        return f"<{self.name!r} {sfields}>"

    def to_serializable(self):
        return (self.name.to_serializable(), self.value.to_serializable())


class BinPtchEntry:
    def __init__(self, hpath, value):
        self.path = BinEntryPath(hpath)
        self.value = value

    def __repr__(self):
        sfields = _repr_indent_list(self.value.fields)
        return f"<BinPtchEntry {self.path!r} {sfields}>"

    def to_serializable(self):
        return self.value.to_serializable()

class BinEntry(BinObjectWithFieldsAndType):
    def __init__(self, hpath, htype, fields):
        self.path = BinEntryPath(hpath)
        super().__init__(htype, fields)

    def __repr__(self):
        sfields = _repr_indent_list(self.fields)
        return f"<BinEntry {self.path!r} {self.type!r} {sfields}>"

class BinFile:
    def __init__(self, f, btype_version=None):
        if isinstance(f, str):
            f = open(f, 'rb')
        magic = f.read(4)
        self.is_patch = magic == b'PTCH'
        if self.is_patch:
            patch_header = struct.unpack('<2L', f.read(8))
            assert patch_header == (1, 0)
            magic = f.read(4)
        if magic != b'PROP':
            raise ValueError("missing magic code")
        reader = BinReader(f, btype_version=btype_version)
        self.version, self.linked_files, entry_types = reader.read_binfile_header()
        self.entries = [reader.read_binfile_entry(htype) for htype in entry_types]
        if self.is_patch and self.version >= 3:
            self.patch_entries = reader.read_patch_section()
        else:
            self.patch_entries = None

    def to_serializable(self):
        serialized = {entry.path.to_serializable(): entry.to_serializable() for entry in self.entries}
        if self.patch_entries is not None:
            serialized["__patches"] = {entry.path.to_serializable(): entry.to_serializable() for entry in self.patch_entries}
        return serialized


class BinReader:
    def __init__(self, f, btype_version=None):
        """
        Initialize a reader for bin files and values

        `btype_version` is a workaround to parse bin types differently
        depending on patch version. Value is based on the patch version.
        """
        self.f = f
        self.btype_version = btype_version or 1008

    def read_fmt(self, fmt):
        length = struct.calcsize(fmt)
        return struct.unpack(fmt, self.f.read(length))

    def read_binfile_header(self):
        """Return a (version, linked_files, entry_types) tuple"""
        version, = self.read_fmt('<L')
        if version >= 2:
            n, = self.read_fmt('<L')
            linked_files = [self.read_string() for _ in range(n)]
        else:
            linked_files = None
        entry_count, = self.read_fmt('<L')
        entry_types = list(self.read_fmt(f"<{entry_count}L"))
        return version, linked_files, entry_types

    def read_binfile_entry(self, htype):
        """Read a single binfile entry"""

        pos = self.f.tell() + 4  # skip 'length' size
        length, hpath, count = self.read_fmt('<LLH')
        values = [self.read_field() for _ in range(count)]
        entry = BinEntry(hpath, htype, values)
        assert self.f.tell() - pos == length
        return entry

    def read_patch_section(self):
        """Reads the entire PTCH section"""

        count = self.read_u32()
        patch_entries = {}
        for _ in range(count):
            hpath = self.read_fmt('<2L')[0]
            btype = self.parse_bintype(self.read_u8())
            objectpath = self.read_string()
            parts = objectpath.split('.')
            binvalue = self._vtype_to_field_reader[btype](self, compute_binhash(parts[-1]), btype)
            if hpath not in patch_entries:
                patch_entries[hpath] = BinPtchEntry(hpath, BinNested([]))

            current_nesting = patch_entries[hpath].value
            for part in parts[:-1]:
                if part not in current_nesting:
                    current_nesting[part] = BinNestedField(compute_binhash(part), BinNested([]))
                current_nesting = current_nesting.getv(part)

            current_nesting[parts[-1]] = binvalue

        return list(patch_entries.values())


    def read_bvalue(self, vtype):
        return self._vtype_to_bvalue_reader[vtype](self)

    def read_empty(self):
        return self.read_fmt('<3H')

    def read_bool(self):
        return self.read_fmt('<?')[0]

    def read_s8(self):
        return self.read_fmt('<b')[0]

    def read_u8(self):
        return self.read_fmt('<B')[0]

    def read_s16(self):
        return self.read_fmt('<h')[0]

    def read_u16(self):
        return self.read_fmt('<H')[0]

    def read_s32(self):
        return self.read_fmt('<i')[0]

    def read_u32(self):
        return self.read_fmt('<I')[0]

    def read_s64(self):
        return self.read_fmt('<q')[0]

    def read_u64(self):
        return self.read_fmt('<Q')[0]

    def read_float(self):
        return self.read_fmt('<f')[0]

    def read_vec2_float(self):
        return self.read_fmt('<2f')

    def read_vec3_float(self):
        return self.read_fmt('<3f')

    def read_vec4_float(self):
        return self.read_fmt('<4f')

    def read_matrix4x4(self):
        return tuple(self.read_fmt('<4f') for _ in range(4))

    def read_rgba(self):
        return self.read_fmt('<4B')

    def read_string(self):
        return self.f.read(self.read_fmt('<H')[0]).decode('utf-8')

    def read_hash(self):
        return BinHashValue(self.read_fmt('<L')[0])

    def read_path(self):
        return BinPathValue(self.read_fmt('<Q')[0])

    def read_link(self):
        return BinEntryPath(self.read_fmt('<L')[0])

    def read_flag(self):
        return self.read_fmt('<B')[0]

    def read_struct(self):
        htype, = self.read_fmt('<L')
        if htype == 0:
            count = 0
        else:
            _, count = self.read_fmt('<LH')
        return BinStruct(htype, [self.read_field() for _ in range(count)])

    def read_embedded(self):
        htype, = self.read_fmt('<L')
        if htype == 0:
            count = 0
        else:
            _, count = self.read_fmt('<LH')
        return BinEmbedded(htype, [self.read_field() for _ in range(count)])

    def read_field(self):
        hname, ftype = self.read_fmt('<LB')
        ftype = self.parse_bintype(ftype)
        return self._vtype_to_field_reader[ftype](self, hname, ftype)

    def read_field_basic(self, hname, btype):
        return BinBasicField(hname, btype, self.read_bvalue(btype))

    def read_field_container(self, hname, btype):
        vtype, _, count = self.read_fmt('<BLL')
        vtype = self.parse_bintype(vtype)
        return BinContainerField(hname, vtype, [self.read_bvalue(vtype) for _ in range(count)])

    def read_field_struct(self, hname, btype):
        return BinStructField(hname, self.read_bvalue(btype))

    def read_field_embedded(self, hname, btype):
        return BinEmbeddedField(hname, self.read_bvalue(btype))

    def read_field_option(self, hname, btype):
        vtype, count = self.read_fmt('<BB')
        assert count in (0, 1)
        vtype = self.parse_bintype(vtype)
        return BinOptionField(hname, vtype, None if count == 0 else self.read_bvalue(vtype))

    def read_field_map(self, hname, btype):
        ktype, vtype, _, count = self.read_fmt('<BBLL')
        ktype, vtype = self.parse_bintype(ktype), self.parse_bintype(vtype)
        # assume key type is hashable
        values = dict((self.read_bvalue(ktype), self.read_bvalue(vtype)) for _ in range(count))
        return BinMapField(hname, ktype, vtype, values)

    def parse_bintype(self, v):
        if self.btype_version < 923:
            if v == 18:
                v = 0x80
            elif v >= 19:
                v = 0x80 + v - 18
        if self.btype_version < 1008:
            if v >= 0x81:
                v += 1
        return BinType(v)


    _vtype_to_bvalue_reader = {
        BinType.EMPTY: read_empty,
        BinType.BOOL: read_bool,
        BinType.S8: read_s8,
        BinType.U8: read_u8,
        BinType.S16: read_s16,
        BinType.U16: read_u16,
        BinType.S32: read_s32,
        BinType.U32: read_u32,
        BinType.S64: read_s64,
        BinType.U64: read_u64,
        BinType.FLOAT: read_float,
        BinType.VEC2_FLOAT: read_vec2_float,
        BinType.VEC3_FLOAT: read_vec3_float,
        BinType.VEC4_FLOAT: read_vec4_float,
        BinType.MATRIX4X4: read_matrix4x4,
        BinType.RGBA: read_rgba,
        BinType.STRING: read_string,
        BinType.HASH: read_hash,
        BinType.PATH: read_path,
        BinType.STRUCT: read_struct,
        BinType.EMBEDDED: read_embedded,
        BinType.LINK: read_link,
        BinType.FLAG: read_flag,
    }

    _vtype_to_field_reader = {
        BinType.EMPTY: read_field_basic,
        BinType.BOOL: read_field_basic,
        BinType.S8: read_field_basic,
        BinType.U8: read_field_basic,
        BinType.S16: read_field_basic,
        BinType.U16: read_field_basic,
        BinType.S32: read_field_basic,
        BinType.U32: read_field_basic,
        BinType.S64: read_field_basic,
        BinType.U64: read_field_basic,
        BinType.FLOAT: read_field_basic,
        BinType.VEC2_FLOAT: read_field_basic,
        BinType.VEC3_FLOAT: read_field_basic,
        BinType.VEC4_FLOAT: read_field_basic,
        BinType.MATRIX4X4: read_field_basic,
        BinType.RGBA: read_field_basic,
        BinType.STRING: read_field_basic,
        BinType.HASH: read_field_basic,
        BinType.PATH: read_field_basic,
        BinType.CONTAINER: read_field_container,
        BinType.CONTAINER2: read_field_container,
        BinType.STRUCT: read_field_struct,
        BinType.EMBEDDED: read_field_embedded,
        BinType.LINK: read_field_basic,
        BinType.OPTION: read_field_option,
        BinType.MAP: read_field_map,
        BinType.FLAG: read_field_basic,
    }


def _to_serializable(v):
    return v.to_serializable() if hasattr(v, 'to_serializable') else v



import os
import re
import zlib
from typing import List, Dict, Union, Optional, Generator, Iterable
from collections import defaultdict
import logging
import requests

from .data import Language
from .tools import write_file_or_remove
from .storage import (
    BaseVersion,
    Storage,
    Patch,
    PatchElement,
    PatchVersion,
    get_system_yaml_version,
    get_exe_version,
)

logger = logging.getLogger(__name__)


class RadsVersion(BaseVersion):
    """Wrapper class for version strings used by RADS

    Solutions and projects all have individual version numbers (e.g. "0.0.1.30").
    The version numbers are actually 32-bit unsigned integers represented using dot-notation, exactly the same as the
    notation used for IPv4 addresses. Notably, each individual number caps at 255, so the version after 0.0.0.255 is
    0.0.1.0.
    """

    def __init__(self, v: Union[str, tuple]):
        super().__init__(v)
        assert len(self.t) == 4, "invalid RADS version format: "


class RadsStorage(Storage):
    """
    Storage based on RADS structure

    Configuration options:
      url -- storage URL (see examples below)
      cdn -- 'default', 'kr' or 'pbe' (incompatible with 'url')

    """

    storage_type = 'rads'

    # all available values are in system.yaml
    # values in use are in RADS/system/system.cfg
    # region is ignored here (it is not actually needed)
    DOWNLOAD_URL = "l3cdn.riotgames.com"
    DOWNLOAD_PATH = "/releases/live"
    DOWNLOAD_PATH_KR = "/KR_CBT"
    DOWNLOAD_PATH_PBE = "/releases/pbe"

    URL_DEFAULT = f"http://{DOWNLOAD_URL}{DOWNLOAD_PATH}/"
    URL_KR = f"http://{DOWNLOAD_URL}{DOWNLOAD_PATH_KR}/"
    URL_PBE = f"http://{DOWNLOAD_URL}{DOWNLOAD_PATH_PBE}/"

    def __init__(self, path, url=None):
        if url is None:
            url = self.URL_DEFAULT
        super().__init__(path, url)

    @classmethod
    def from_conf_data(cls, conf):
        if 'cdn' in conf:
            if 'url' in conf:
                raise ValueError("'url' and 'cdn' are mutually exclusive")
            url = getattr(cls, f"URL_{conf['cdn']}".upper())
        else:
            url = conf.get('url')
        return cls(conf['path'], url)

    def list_projects(self) -> List['RadsProject']:
        """List projects present in storage"""
        ret = []
        base = self.fspath("projects")
        for name in os.listdir(base):
            if os.path.isdir(f"{base}/{name}/releases"):
                ret.append(RadsProject(self, name))
        return ret

    def list_solutions(self) -> List['RadsSolution']:
        """List solutions present in storage"""
        ret = []
        base = self.fspath("solutions")
        for name in os.listdir(base):
            if os.path.isdir(f"{base}/{name}/releases"):
                ret.append(RadsSolution(self, name))
        return ret

    def patch_elements(self, stored=False):
        solution_names = ('league_client_sln', 'lol_game_client_sln')

        # peek next element for each solution
        class Peeker:
            def __init__(self, it):
                self.it = it
                self.cur = None

            def peek(self):
                if self.cur is None:
                    try:
                        self.cur = next(self.it)
                    except StopIteration:
                        pass
                return self.cur

            def consume(self):
                assert self.cur is not None
                self.cur = None

        # drop solution versions without a patch
        # convert them to patch elements
        def gen_solution_elements(name):
            solution = RadsSolution(self, name)
            for sv in solution.versions(stored=stored):
                patch = sv.patch_version()
                if patch is None:
                    continue
                yield RadsPatchElement(sv)

        # for each solution, peek the next elements to yield the highest version
        peekers = [Peeker(gen_solution_elements(name)) for name in solution_names]
        while True:
            best_peeker, best_elem = None, None
            for peeker in peekers:
                elem = peeker.peek()
                if elem is None:
                    continue
                if best_elem is None or elem.version > best_elem.version:
                    best_peeker, best_elem = peeker, elem
            if best_peeker is None:
                break  # exhausted
            yield best_elem
            best_peeker.consume()


class RadsSolution:
    """A Solution has multiple versions and contains many Projects.

    The Riot Application Distribution System (RADS) has two Solutions: `league_client_sln` and `lol_game_client_sln`.
    The 'league_client_sln' contains data the client (LCU), and the `lol_game_client_sln` contains data for the game client.
    These classes will likely work with other solutions, although some functionality may need to be extended.

    There are multiple versions of a given solution, which can be accessed via the `.versions()` method.
    All versions of a solution can be downloaded and extracted via the `.download()` method.

    Each version of a solution contains multiple projects pertaining to different locales.
    """

    def __init__(self, storage: RadsStorage, name):
        self.storage = storage
        self.path = f"solutions/{name}/releases"
        self.name = name

    def __str__(self):
        return f"rads:{self.name}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.name}>"

    def __eq__(self, other):
        if isinstance(other, RadsSolution):
            return self.name == other.name
        return False

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        if isinstance(other, RadsSolution):
            return self.name < other.name
        return NotImplemented

    def versions(self, stored=False) -> List['RadsSolutionVersion']:
        """Retrieve a sorted list of versions of this solution

        If stored is True, only versions in storage are used (to avoid downloading new files).
        """

        if stored:
            fspath = self.storage.fspath(self.path)
            if not os.path.isdir(fspath):
                return []  # solution not in storage
            listing = []
            for path in os.listdir(fspath):
                if not os.path.isdir(os.path.join(fspath, path)):
                    continue
                listing.append(path)
        else:
            logger.debug(f"retrieve versions of {self}")
            listing = self.storage.request_text(f"{self.path}/releaselisting").splitlines()
        return sorted(RadsSolutionVersion(self, RadsVersion(l)) for l in listing)

    def download(self, langs):
        for v in self.versions():
            v.download(langs)


class RadsSolutionVersion:
    """A single version of a RadsSolution.

    Each RadsSolutionVersion contains data for multiple projects, accessible via the `RadsSolutionVersion.projects` method.
    There is one "main" project, and one project for each language.

    The data contained in a RadsSolutionVersion can be downloaded and extracted via the `.download()` method.
    """

    def __init__(self, solution: RadsSolution, version: 'RadsVersion'):
        self.path = f"{solution.path}/{version}"
        self.solution = solution
        self.version = version

    def __str__(self):
        return f"{self.solution}={self.version}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.solution.name}={self.version}>"

    def __eq__(self, other):
        if isinstance(other, RadsSolutionVersion):
            return self.solution == other.solution and self.version == other.version
        return False

    def __hash__(self):
        return hash((self.solution, self.version))

    def __lt__(self, other):
        if isinstance(other, RadsSolutionVersion):
            if self.solution < other.solution:
                return True
            elif self.solution == other.solution:
                return self.version > other.version
            else:
                return False
        return NotImplemented

    def dependencies(self) -> Dict[Union[Language, None], List['RadsProjectVersion']]:
        """Parse dependencies from the solutionmanifest

        Return a map of project versions for each language.
        The entry None is set to all required project versions.
        """

        logger.debug(f"retrieve dependencies of {self}")

        path = f"{self.path}/solutionmanifest"
        self.solution.storage.download(path, path)
        with open(self.solution.storage.fspath(path)) as f:
            lines = f.read().splitlines()
        assert lines[0] == "RADS Solution Manifest", "unexpected solutionmanifest magic line"
        assert lines[1] == "1.0.0.0", "unexpected solutionmanifest version"
        assert lines[2] == self.solution.name, "solution name mismatch in solutionmanifest header"
        assert lines[3] == self.version, "solution version mismatch in solutionmanifest header"
        idx = 4

        required_projects = []  # [name, ...]
        projects = {}  # {name: RadsProjectVersion}
        nprojects, idx = int(lines[idx]), idx + 1
        for _ in range(nprojects):
            (name, version, unk1, unk2), idx = lines[idx:idx+4], idx + 4
            unk1, unk2 = int(unk1), int(unk2)
            if unk1 == 0:
                required_projects.append(name)
            else:
                assert unk1 == 10
            assert unk2 == 0
            projects[name] = RadsProjectVersion(RadsProject(self.solution.storage, name), RadsVersion(version))

        langs = {}  # {Language: [RadsProjectVersion, ...]}
        nlangs, idx = int(lines[idx]), idx + 1
        for _ in range(nlangs):
            (lang, unk1, ndeps), idx = lines[idx:idx+3], idx + 3
            unk1, ndeps = int(unk1), int(ndeps)
            assert unk1 == 0
            deps, idx = lines[idx:idx+ndeps], idx + ndeps
            langs[Language(lang)] = [projects[name] for name in deps]

        langs[None] = list(projects[name] for name in required_projects)
        return langs

    def projects(self, langs=True) -> List['RadsProjectVersion']:
        """Return a list of projects for provided language(s)"""
        dependencies = self.dependencies()
        if langs is False:
            return dependencies[None]
        elif langs is True:
            return list({pv for pvs in dependencies.values() for pv in pvs})
        elif isinstance(langs, Language):
            return dependencies[langs]
        else:
            return list({pv for lang in langs for pv in dependencies[lang]})

    def filepaths(self, langs) -> Generator[str, None, None]:
        """Generate the extract path of files in the solution version"""
        for pv in self.projects(langs):
            yield from pv.filepaths()

    def download(self, langs=True):
        """Download solution version files"""

        logger.info(f"downloading solution {self}")
        for pv in self.projects(langs):
            pv.download()

    def patch_version(self) -> Optional[PatchVersion]:
        """Return patch version or None if there is None

        This method reads/writes version from/to cache.
        """

        # for PBE: version is always "main"
        if self.solution.storage.url == RadsStorage.URL_PBE:
            return PatchVersion("main")

        cache = self.solution.storage.fspath(f"{self.path}/_patch_version")
        if os.path.isfile(cache):
            logger.debug(f"retrieving patch version for {self} from cache")
            with open(cache) as f:
                version = f.read().strip()
                version = PatchVersion(version) if version else None
        else:
            version = self._retrieve_patch_version()
            if version is None:
                logger.warning(f"failed to retrieve patch version for {self}")
            else:
                with open(cache, 'w') as f:
                    f.write(f"{version}\n")
        return version

    def _retrieve_patch_version(self) -> Optional[PatchVersion]:
        """Retrieve patch version from game files (no cache handling)

        Return None if there is no patch version (because files are not
        available anymore on Riot's CDN).
        Raise an exception if patch version cannot be retrieved.
        """

        logger.debug(f"retrieving patch version for {self}")

        retrievers = {
            # solution_name: (project_name, file_name, extractor)
            'league_client_sln': (
                'league_client',
                'system.yaml',
                get_system_yaml_version,
            ),
            'lol_game_client_sln': (
                'lol_game_client',
                'League of Legends.exe',
                get_exe_version,
            ),
        }

        try:
            project_name, file_name, extractor = retrievers[self.solution.name]
        except KeyError:
            raise RuntimeError(f"no known way to retrieve patch version for solution {self.solution.name}")

        for pv in self.projects(False):
            if pv.project.name == project_name:
                break
        else:
            raise ValueError(f"{project_name} project not found for {self}")

        try:
            filepaths = pv.filepaths()
        except requests.exceptions.HTTPError as e:
            # some packagemanifest files are not available anymore
            # for these project versions, there is no patch version
            if e.response is not None and e.response.status_code == 404:
                return None
            raise

        path_suffix = f'/{file_name}'
        for path in filepaths:
            if path.endswith(path_suffix):
                fspath = self.solution.storage.fspath(path)
                if not os.path.isfile(path):
                    pv.extract([path])
                break
        else:
            # packagemanifest for league_client<=0.0.0.43 doesn't alway contain system.yaml
            if pv.project.name == 'league_client' and pv.version <= RadsVersion('0.0.0.43'):
                return None
            raise ValueError(f"'{file_name}' not found for {pv}")

        version = extractor(fspath)
        return PatchVersion(version)


class RadsProject:
    """A RadsProject is a subset of data for a specific locale, or the data for the main/default/common locale.

    There are multiple versions of a given project, which can be accessed via the `.versions()` method.
    All versions of the project can be downloaded and extracted via the `.download()` method.
    The data in ProjectVersions are contained in Bin files, which are extracted.
    """

    def __init__(self, storage: RadsStorage, name):
        self.storage = storage
        self.path = f"projects/{name}/releases"
        self.name = name

    def __str__(self):
        return f"rads:{self.name}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.name}>"

    def __eq__(self, other):
        if isinstance(other, RadsProject):
            return self.name == other.name
        return False

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        if isinstance(other, RadsProject):
            return self.name < other.name
        return NotImplemented

    def versions(self) -> List['RadsProjectVersion']:
        """Retrieve the list of versions of this project"""
        logger.debug(f"retrieve versions of {self}")
        listing = self.storage.request_text(f"{self.path}/releaselisting")
        return [RadsProjectVersion(self, RadsVersion(l)) for l in listing.splitlines()]

    def download(self):
        for v in self.versions():
            v.download()


class RadsProjectVersion:
    """A single version of a RadsProject.

    The data contained in a project can be downloaded and extracted via the `.download()` method.
    The data in these ProjectVersions are contained in Bin files, which are extracted.
    """

    def __init__(self, project: RadsProject, version: 'Version'):
        self.path = f"{project.path}/{version}"
        self.project = project
        self.version = version
        self._package_files = None  # {extract_path: BinPackageFile}

    def __str__(self):
        return f"{self.project}={self.version}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.project.name}={self.version}>"

    def __eq__(self, other):
        if isinstance(other, RadsProjectVersion):
            return self.project == other.project and self.version == other.version
        return False

    def __hash__(self):
        return hash((self.project, self.version))

    def __lt__(self, other):
        if isinstance(other, RadsProjectVersion):
            if self.project < other.project:
                return True
            elif self.project == other.project:
                return self.version > other.version
            else:
                return False
        return NotImplemented

    def _get_package_files(self) -> Dict[str, 'BinPackageFile']:
        """Retrieve files from packagemanifest"""

        if self._package_files is None:
            manifest_path = f"{self.path}/packagemanifest"
            manifest_urlpath = f"{self.path}/packages/files/packagemanifest"
            self.project.storage.download(manifest_urlpath, manifest_path)
            files = BinPackageFile.from_package_manifest(self.project.storage.fspath(manifest_path))
            self._package_files = {pf.extract_path: pf for pf in files}
        return self._package_files

    def filepaths(self) -> Dict[str, 'BinPackageFile']:
        """Generate the extract path of files in the project version"""
        return self._get_package_files()

    def extract(self, paths=None):
        """Download packages and extract files

        A subset of paths to extract can be provided (they must exist in the project version).
        """

        all_files = self._get_package_files()
        if paths is None:
            extracted_files = all_files.values()
        else:
            extracted_files = [all_files[path] for path in paths]

        # filter already extracted file
        extracted_files = [pf for pf in extracted_files if not os.path.isfile(self.project.storage.fspath(pf.extract_path))]

        # group files by package
        files_by_package = defaultdict(list)
        for pf in extracted_files:
            files_by_package[pf.package].append(pf)

        package_files_path = f"{self.path}/packages/files"

        for package, files in files_by_package.items():
            with self.project.storage.stream(f"{package_files_path}/{package}") as reader:
                # sort files by offset to extract while streaming the bin file
                for pkgfile in sorted(files, key=lambda f: f.offset):
                    logger.debug(f"extracting {pkgfile.path}")
                    reader.skip_to(pkgfile.offset)
                    fspath = self.project.storage.fspath(pkgfile.extract_path)
                    with write_file_or_remove(fspath) as fout:
                        if pkgfile.compressed:
                            zobj = zlib.decompressobj(zlib.MAX_WBITS | 32)
                            def writer(data):
                                return fout.write(zobj.decompress(data))
                            reader.copy(writer, pkgfile.size)
                            fout.write(zobj.flush())
                        else:
                            reader.copy(fout.write, pkgfile.size)

    def download(self):
        """Download project version files"""
        logger.info(f"downloading project {self}")
        self.project.storage.download(f"{self.path}/releasemanifest", None)
        self.extract()


class RadsPatchElement(PatchElement):
    """Patch element from a RADS storage"""

    solution_name_to_element_name = {
        'league_client_sln': 'client',
        'lol_game_client_sln': 'game',
    }

    def __init__(self, sv: RadsSolutionVersion):
        name = self.solution_name_to_element_name[sv.solution.name]
        version = sv.patch_version()
        if version is None:
            raise ValueError(f"unknown patch version for {sv}")
        super().__init__(name, version)
        self.solution_version = sv

    def download(self, langs=True):
        self.solution_version.download(langs=langs)

    def fspaths(self, langs=True):
        sv = self.solution_version
        storage = sv.solution.storage
        return (storage.fspath(path) for path in sv.filepaths(langs=langs))

    def relpaths(self, langs=True):
        sv = self.solution_version
        return (path.split('/', 5)[5].lower() for path in sv.filepaths(langs=langs))

    def paths(self, langs=True):
        sv = self.solution_version
        storage = sv.solution.storage
        for path in sv.filepaths(langs=langs):
            yield (storage.fspath(path), path.split('/', 5)[5].lower())


class BinPackageFile:
    """A single file in a BIN package"""

    __slots__ = ('path', 'package', 'offset', 'size', 'compressed', 'extract_path')

    def __init__(self, line):
        path, self.package, offset, size, typ = line.split(',')
        self.path = path[1:]  # remove leading '/'
        self.offset = int(offset)
        self.size = int(size)
        self.compressed = self.path.endswith('.compressed')
        if self.compressed:
            self.extract_path = self.path[:-11]  # remove the '.compressed' suffix
        else:
            self.extract_path = self.path

    def __str__(self):
        return f"<{self.__class__.__name__} {self.path!r}>"

    @classmethod
    def from_package_manifest(cls, path) -> Generator['BinPackageFile', None, None]:
        with open(path) as f:
            line = f.readline()
            assert line.startswith('PKG1'), "unexpected packagemanifest magic line"
            for line in f:
                yield cls(line)


def parse_rads_component(storage: RadsStorage, component: str):
    """Parse a component string representation to an object"""

    m = re.match(r'^(?:([sp]):)?(\w+)(?:=(|[0-9]+(?:\.[0-9]+)*|main)?)?$', component)
    if not m:
        raise ValueError(f"invalid component: {component}")
    typ, name, version = m.group(1, 2, 3)
    if not typ:
        if name == 'patch':
            typ = 'patch'
        elif name.endswith('_sln'):
            typ = 's'
        else:
            typ = 'p'

    if typ == 'p':
        project = RadsProject(storage, name)
        if version is None:
            return project
        elif version == '':
            return project.versions()[0]
        else:
            return RadsProjectVersion(project, RadsVersion(version))
    elif typ == 's':
        solution = RadsSolution(storage, name)
        if version is None:
            return solution
        elif version == '':
            return solution.versions()[0]
        else:
            return RadsSolutionVersion(solution, RadsVersion(version))
    elif typ == 'patch':
        if version is None:
            raise ValueError(f"patch requires a version")
        elif version == '':
            return storage.patch(None)
        else:
            return storage.patch(version)

